# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the base PhaseDiagram creator that interfaces
with QueryConvexHull and EnsembleHull.

"""


from traceback import print_exc
import bisect

import scipy.spatial
import numpy as np

from matador.utils.hull_utils import (
    barycentric2cart, vertices2plane, vertices2line, FakeHull, is_point_in_triangle
)
from matador.utils.chem_utils import get_formula_from_stoich
from matador.utils.cursor_utils import get_array_from_cursor, display_results, set_cursor_from_array

EPS = 1e-12


class PhaseDiagram:
    """ This class encapsulates the actual phase data, e.g. the actual
    energy and compositions found to be stable.

    Attributes:
        structures (numpy.ndarray): the array passed to init used to
            make the hull, with the first (num_species-1) columns
            containing normalised concentrations, and the final column
            containing formation energy.
        convex_hull (scipy.spatial.ConvexHull): the actual convex hull
            returned by SciPy.
        formation_key (list): index/key specification of formation energy
            per atom from top level of each document.

    """
    def __init__(self, cursor, formation_key, dimension):
        """ Compute the convex hull of data passed, to retrieve hull
        distances and thus stable structures.

        Parameters:
            cursor (list[dict]): list of matador documents to make
                phase diagram from.
            formation_key (str or list):  location of the formation energy
                inside each document, either a single key or iterable of
                keys to use with `recursive_get`.

        """
        self._dimension = dimension
        self.cursor = cursor
        self.formation_key = formation_key

        structures = np.hstack((
            get_array_from_cursor(cursor, 'concentration').reshape(len(cursor), dimension-1),
            get_array_from_cursor(cursor, self.formation_key).reshape(len(cursor), 1)))

        # define self._structure_slice as the filtered array of points actually used to create the convex hull
        # which can include/exclude points from the passed structures. This array is the one indexed by
        # vertices/simplices in ConvexHull

        if self._dimension == 3:
            # add a point "above" the hull
            # for simple removal of extraneous vertices (e.g. top of 2D hull)
            dummy_point = [0.333, 0.333, 1e5]
            # if ternary, use all structures, not just those with negative eform for compatibility reasons
            self._structure_slice = np.vstack((structures, dummy_point))
        else:
            # filter out those with positive formation energy, to reduce expense computing hull
            self._structure_slice = structures[np.where(structures[:, -1] <= 0 + EPS)]

        # filter out "duplicates" in _structure_slice
        # this prevents breakages if no structures are on the hull and chempots are duplicated
        # but it might be faster to hardcode this case individually
        self._structure_slice = np.unique(self._structure_slice, axis=0)

        # if we only have the chempots (or worse) with negative formation energy, don't even make the hull
        if len(self._structure_slice) <= dimension:
            if len(self._structure_slice) < dimension:
                raise RuntimeError('No chemical potentials on hull... either mysterious use of custom chempots, or worry!')
            self.convex_hull = FakeHull()
        else:
            try:
                self.convex_hull = scipy.spatial.ConvexHull(self._structure_slice)
            except scipy.spatial.qhull.QhullError:
                print(self._structure_slice)
                print('Error with QHull, plotting formation energies only...')
                print_exc()
                self.convex_hull = FakeHull()

        # remove vertices that have positive formation energy
        filtered_vertices = [vertex for vertex in self.convex_hull.vertices if self._structure_slice[vertex, -1] <= 0 + EPS]
        bad_simplices = set()
        for ind, simplex in enumerate(self.convex_hull.simplices):
            for vertex in simplex:
                if vertex not in filtered_vertices:
                    bad_simplices.add(ind)

        filtered_simplices = [simplex for ind, simplex in enumerate(self.convex_hull.simplices) if ind not in bad_simplices]

        self.convex_hull = FakeHull()
        self.convex_hull.points = self._structure_slice
        self.convex_hull.vertices = list(filtered_vertices)
        self.convex_hull.simplices = list(filtered_simplices)

        self.hull_dist = self.get_hull_distances(structures, precompute=True)
        set_cursor_from_array(self.cursor, self.hull_dist, 'hull_distance')
        self.structures = structures
        self.stable_structures = [doc for doc in self.cursor if doc['hull_distance'] < EPS]

    def __str__(self):
        """ Print underlying phase diagram. """
        return display_results(self.cursor,
                               hull=True,
                               colour=False,
                               energy_key=self.formation_key,
                               sort=False,
                               return_str=True)

    def get_hull_distances(self, structures, precompute=False, **kwargs):
        """ Returns array of distances to pre-computed binary or ternary
        hull, from array containing concentrations and energies.

        Parameters:
            structures (numpy.ndarray): N x n array of concentrations and
                enthalpies for N structures, with up to 2 columns of
                concentrations and the last column containing the
                structure's formation enthalpy.

        Keyword arguments:
            precompute (bool): whether or not to bootstrap hull
                distances from previously computed values at the same
                stoichiometry.

        Returns:
            numpy.ndarray: N-dim array storing distances to
                the hull for N structures,

        """

        if precompute:
            # dict with formula keys, containing tuple of pre-computed enthalpy/atom and hull distance
            cached_formula_dists = dict()
            cache_hits = 0
            cache_misses = 0

        if isinstance(structures, list):
            structures = np.asarray(structures)

        # if only chem pots on hull, dist = energy
        if len(self._structure_slice) == self._dimension:
            hull_dist = np.ones((len(structures)))
            hull_dist = structures[:, -1]

        # if binary hull, do binary search
        elif self._dimension == 2:
            tie_line_comp = self._structure_slice[self.convex_hull.vertices, 0]
            tie_line_energy = self._structure_slice[self.convex_hull.vertices, -1]
            tie_line_comp = np.asarray(tie_line_comp)
            tie_line_energy = tie_line_energy[np.argsort(tie_line_comp)]
            tie_line_comp = tie_line_comp[np.argsort(tie_line_comp)]

            hull_dist = np.empty((len(structures)))
            hull_dist.fill(np.nan)
            if precompute:
                for ind, _ in enumerate(structures):
                    formula = get_formula_from_stoich(self.cursor[ind]['stoichiometry'], sort=True, tex=False)
                    if formula in cached_formula_dists:
                        hull_dist[ind] = (structures[ind, -1] - cached_formula_dists[formula][0] +
                                          cached_formula_dists[formula][1])
                        cache_hits += 1
                    else:
                        i = bisect.bisect_left(tie_line_comp, structures[ind, 0])
                        gradient, intercept = vertices2line([[tie_line_comp[i-1], tie_line_energy[i-1]],
                                                             [tie_line_comp[i], tie_line_energy[i]]])
                        # calculate hull_dist
                        hull_dist[ind] = structures[ind, -1] - (gradient * structures[ind, 0] + intercept)
                        cached_formula_dists[formula] = (structures[ind, -1], hull_dist[ind])
                        cache_misses += 1
            else:
                for ind, _ in enumerate(structures):
                    i = bisect.bisect_left(tie_line_comp, structures[ind, 0])
                    gradient, intercept = vertices2line([[tie_line_comp[i-1], tie_line_energy[i-1]],
                                                         [tie_line_comp[i], tie_line_energy[i]]])
                    # calculate hull_dist
                    hull_dist[ind] = structures[ind, -1] - (gradient * structures[ind, 0] + intercept)

        # if ternary, use barycentric coords
        elif self._dimension == 3:
            # loop through structures and find which plane they correspond to
            # using barycentric coordinates, if a formula has already been
            # computed then calculate delta relative to that and skip
            self.convex_hull.planes = [[self._structure_slice[vertex] for vertex in simplex]
                                       for simplex in self.convex_hull.simplices]
            structures_finished = [False] * len(structures)
            hull_dist = np.empty(len(structures))
            hull_dist.fill(np.nan)
            cart_planes_inv = []
            planes_height_fn = []
            for ind, plane in enumerate(self.convex_hull.planes):
                cart_planes = barycentric2cart(plane).T
                cart_planes[-1, :] = 1
                # if projection of triangle in 2D is a line, do binary search
                if np.linalg.det(cart_planes) == 0:
                    cart_planes_inv.append(None)
                    planes_height_fn.append(None)
                else:
                    cart_planes_inv.append(np.linalg.inv(cart_planes))
                    planes_height_fn.append(vertices2plane(plane))
            for idx, structure in enumerate(structures):
                for ind, plane in enumerate(self.convex_hull.planes):
                    if cart_planes_inv[ind] is None:
                        continue
                    if precompute and get_formula_from_stoich(self.cursor[idx]['stoichiometry'], sort=True,
                                                              tex=False) in cached_formula_dists:
                        formula = get_formula_from_stoich(self.cursor[idx]['stoichiometry'], sort=True, tex=False)
                        if formula in cached_formula_dists:
                            cache_hits += 1
                            hull_dist[idx] = (structures[idx, -1] - cached_formula_dists[formula][0] +
                                              cached_formula_dists[formula][1])
                            structures_finished[idx] = True

                    elif is_point_in_triangle(structure, cart_planes_inv[ind], preprocessed_triangle=True):
                        structures_finished[idx] = True
                        hull_dist[idx] = planes_height_fn[ind](structure)
                        if precompute:
                            cached_formula_dists[
                                get_formula_from_stoich(self.cursor[idx]['stoichiometry'], sort=True,
                                                        tex=False)] = (structure[-1], hull_dist[idx])
                            cache_misses += 1
                        break

            # mask values very close to 0 with 0
            hull_dist[np.where(np.abs(hull_dist) < EPS)] = 0

            failed_structures = []
            for ind, structure in enumerate(structures_finished):
                if not structure:
                    failed_structures.append(ind)

            if failed_structures:
                raise RuntimeError('There were issues calculating the hull distance for {} structures.'
                                   .format(len(failed_structures)))

        # otherwise, set to zero until proper N-d distance can be implemented
        else:
            raise NotImplementedError
            # self.hull.planes = [[self._structure_slice[vertex] for vertex in simplex]
            #                     for simplex in self.hull.simplices]
            # for idx, structure in enumerate(structures):
            #     if precompute and get_formula_from_stoich(self.cursor[idx]['stoichiometry'],
            #                                               tex=False) in cached_formula_dists:
            #         formula = get_formula_from_stoich(self.cursor[idx]['stoichiometry'], tex=False)
            #         if formula in cached_formula_dists:
            #             cache_hits += 1
            #             hull_dist[idx] = (structures[idx, -1] - cached_formula_dists[formula][0] +
            #                               cached_formula_dists[formula][1])
            #         structures_finished[idx] = True
            #     else:
            #         if point_in_polygon(structure[:-1], plane):
            #             hull_dist[idx] = 0
            #             continue

            # loop over compositions
            #   loop over planes
            #       check if point is interior to plane
            #           if so, compute "height" above plane, using equation of hyperplane from scipy
        if np.isnan(hull_dist).any():
            raise RuntimeError("Some hull distances failed")

        return hull_dist

    def point_in_polygon(self):
        """ Use "ray-tracing" or winding number approach
        to compute whether a point lies inside a polygon.
        """
        raise NotImplementedError



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module contains functionality to update and overwrite database
entries with specific tasks, e.g. symmetry and substructure analysis.
This can be especially useful for "patching" the data in databases
created with older matador versions.

"""


from traceback import print_exc
import os

import pymongo as pm

from matador.utils.print_utils import print_notify, print_warning, print_failure


class Refiner:
    """ Refiner implements methods to alter certain parts of the
    database in place, either in overwrite, set or compare/display mode.
    Current modifiables are space groups, substructures,
    the set of elements, tags and DOIs.

    """
    def __init__(self, cursor, collection=None, task=None, mode='display', **kwargs):
        """ Parses args and initiates modification.

        Parameters:
            cursor (list of dicts): matador cursor to refine.

        Keyword arguments:
            collection (Collection): mongodb collection to query/edit.
            task (str/callable): one of 'sym', 'spg', 'elem_set', 'tag', 'doi' or 'source',
                or a custom function that takes in and returns a cursor and field to modify.
            mode (str): one of 'display', 'overwrite', 'set'.

        """
        possible_tasks = ['sym', 'spg', 'elem_set', 'tag', 'doi', 'source', 'pspot', 'raw']
        possible_modes = ['display', 'overwrite', 'set']

        if mode not in possible_modes:
            print('Mode not understood, defaulting to "display".')
            mode = 'display'
        if collection is None and mode in ['overwrite', 'set']:
            raise SystemExit('Impossible to overwite or set without db collection, exiting...')
        if task is None:
            raise SystemExit('No specified task, exiting...')
        if task not in possible_tasks and not callable(task):
            raise SystemExit('Did not understand task, please choose one of ' + ', '.join(possible_tasks))
        if task == 'tag' and mode == 'set':
            raise SystemExit('Task "tags" and mode "set" will not alter the database, please use mode "overwrite".')

        self.cursor = list(cursor)
        self.diff_cursor = []
        self.collection = collection
        self.mode = mode
        self.changed_count = 0
        self.failed_count = 0
        self.args = kwargs

        if task in ['spg', 'sym']:
            if kwargs.get('symprec'):
                self.symmetry(symprec=kwargs.get('symprec'))
            else:
                self.symmetry()
            self.field = 'space_group'
        elif task == 'elem_set':
            self.elem_set()
            self.field = 'elems'
        elif task == 'tag':
            self.field = 'tags'
            self.tag = self.args.get('new_tag')
            if self.tag is None:
                print_warning('No new tag defined, nothing will be done.')
            else:
                self.add_tag()
        elif task == 'doi':
            self.field = 'doi'
            self.doi = self.args.get('new_doi')
            if self.doi is None:
                print_warning('No new DOI defined, nothing will be done.')
            else:
                self.add_doi()
        elif task == 'source':
            self.field = 'root_source'
            self.add_root_source()
        elif task == 'pspot':
            self.field = 'species_pot'
            self.tidy_pspots()
        elif task == 'raw':
            self.field = '_raw'
            self.add_raw_data()
        elif callable(task):
            print('Using custom task function: {}'.format(task))
            self.diff_cursor, self.field = task(self.cursor)

        self.changed_count = len(self.diff_cursor)
        print(self.changed_count, '/', len(self.cursor), 'to be changed.')
        print(self.failed_count, '/', len(self.cursor), 'failed.')

        if self.mode in ['set', 'overwrite'] and self.changed_count > 0:
            self.update_docs()

    def update_docs(self):
        """ Updates documents in database with correct priority. """
        requests = []
        # if in "set" mode, do not overwrite, just apply
        if self.mode == 'set':
            for _, doc in enumerate(self.diff_cursor):
                requests.append(pm.UpdateOne({'_id': doc['_id'], self.field: {'$exists': False}},
                                             {'$set': {self.field: doc[self.field]}}))
        # else if in overwrite mode, overwrite previous field
        elif self.mode == 'overwrite':
            for _, doc in enumerate(self.diff_cursor):
                requests.append(pm.UpdateOne({'_id': doc['_id']}, {'$set': {self.field: doc[self.field]}}))
        if self.args.get('debug'):
            for request in requests:
                print(request)

        result = self.collection.bulk_write(requests)
        print_notify(str(result.modified_count) + ' docs modified.')

    def symmetry(self, symprec=1e-3):
        """ Compute space group with spglib. """
        from matador.utils.cell_utils import doc2spg
        import spglib as spg
        print('Refining symmetries...')
        if self.mode == 'display':
            print_warning('{}'.format('At symprec: ' + str(symprec)))
            print_warning("{:^36}{:^16}{:^16}".format('text_id', 'new sg', 'old sg'))
        for _, doc in enumerate(self.cursor):
            try:
                spg_cell = doc2spg(doc)
                sg = spg.get_spacegroup(spg_cell, symprec=symprec).split(' ')[0]
                if sg != doc['space_group']:
                    self.changed_count += 1
                    self.diff_cursor.append(doc)
                    if self.mode == 'display':
                        print_notify("{:^36}{:^16}{:^16}"
                                     .format(doc['text_id'][0]+' '+doc['text_id'][1], sg, doc['space_group']))
                    doc['space_group'] = sg
                else:
                    if self.mode == 'display':
                        print("{:^36}{:^16}{:^16}"
                              .format(doc['text_id'][0]+' '+doc['text_id'][1], sg, doc['space_group']))
            except Exception:
                self.failed_count += 1
                if self.args.get('debug'):
                    print_exc()
                    print_failure('Failed for' + ' '.join(doc['text_id']))

    def elem_set(self):
        """ Imbue documents with the set of elements,
        i.e. set(doc['atom_types']), for quicker look-up.

        """
        for doc in self.cursor:
            try:
                doc['elems'] = list(set(doc['atom_types']))
                self.diff_cursor.append(doc)
                self.changed_count += 1
            except Exception as oops:
                if self.args.get('debug'):
                    print(repr(oops))
                self.failed_count += 1

    def add_tag(self):
        """ Add a tag to each document. """
        for _, doc in enumerate(self.cursor):
            try:
                if 'tags' in doc:
                    if doc['tags'] is None:
                        doc['tags'] = list()
                    else:
                        doc['tags'] = list(doc['tags'])
                    doc['tags'].append(self.tag)
                else:
                    doc['tags'] = [self.tag]
                self.diff_cursor.append(doc)
                self.changed_count += 1
            except Exception as error:
                print(repr(error))
                self.failed_count += 1

    def add_doi(self):
        """ Add a doi to each document. """
        if self.doi.count('/') != 1:
            raise SystemExit('Malformed DOI... please use xxxxx/xxxxx format.')
        for _, doc in enumerate(self.cursor):
            try:
                if 'doi' in doc:
                    doc['doi'].append(self.doi)
                else:
                    doc['doi'] = [self.doi]
                self.diff_cursor.append(doc)
                self.changed_count += 1
            except Exception as error:
                print(repr(error))
                self.failed_count += 1

    def add_root_source(self):
        """ Add the "root_source" key to a document in the database,
        i.e. the name of the structure, minus file extension.

        """
        from matador.utils.chem_utils import get_root_source
        for _, doc in enumerate(self.cursor):
            try:
                if 'root_source' in doc:
                    continue
                else:
                    doc['root_source'] = get_root_source(doc['source'])
                    self.diff_cursor.append(doc)
                    self.changed_count += 1
            except Exception as error:
                print(repr(error))
                self.failed_count += 1

    def tidy_pspots(self):
        """ Loop over all documents and make sure they only have pspots
        for the elements that exist in the structure.

        """
        for _, doc in enumerate(self.cursor):
            try:
                marked_for_del = []
                atoms = set(doc['atom_types'])
                for elem in doc['species_pot']:
                    if elem not in atoms:
                        marked_for_del.append(elem)
                for elem in marked_for_del:
                    del doc['species_pot'][elem]
                    self.diff_cursor.append(doc)
                    self.changed_count += 1
            except Exception as error:
                print(repr(error))
                self.failed_count += 1

    def add_raw_data(self):
        """ Loop over all documents in the query and try to open the files
        listed under their `source` fields, storing them under the `_raw` key.

        """

        for _, doc in enumerate(self.cursor):
            try:
                sources = doc['source']
                raw_files = []
                for source in sources:
                    if os.path.isfile(source):
                        with open(source, 'r') as f:
                            raw_files.append((source, f.readlines()))

                doc['_raw'] = raw_files
                self.diff_cursor.append(doc)
                self.changed_count += 1
            except Exception as error:
                print(repr(error))
                self.failed_count += 1



from pathlib import Path

with open(Path(__file__).parent.parent.joinpath('data/words').resolve(), 'r') as f:
    WORDS = f.readlines()
with open(Path(__file__).parent.parent.joinpath('data/nouns').resolve(), 'r') as f:
    NOUNS = f.readlines()

__all__ = ['NOUNS', 'WORDS']



#!/usr/bin/env python
import unittest
import os
import numpy as np
from matador.utils.cell_utils import (
    abc2cart,
    cart2abc,
    cart2volume,
    create_simple_supercell,
)
from matador.utils.cell_utils import cart2frac, frac2cart
from matador.utils.cell_utils import doc2spg, cart2abcstar, real2recip
from matador.scrapers.castep_scrapers import (
    castep2dict,
    res2dict,
    cell2dict,
    bands2dict,
)
from matador.fingerprints.pdf import PDF, PDFOverlap
from matador.export import doc2cell
from matador.crystal import Crystal

VERBOSITY = 0

try:
    from matador.utils.cell_utils import get_seekpath_kpoint_path
    from seekpath import get_path

    IMPORTED_SEEKPATH = True
except Exception:
    IMPORTED_SEEKPATH = False

# grab abs path for accessing test data
REAL_PATH = "/".join(os.path.realpath(__file__).split("/")[:-1]) + "/"


class CellUtilTest(unittest.TestCase):
    """ Tests cell util functions. """

    def test_cart2abc(self):
        castep_fname = REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep"
        self.assertTrue(os.path.isfile(castep_fname))
        test_doc, s = castep2dict(castep_fname, db=True, verbosity=VERBOSITY)
        try:
            self.assertTrue(
                np.allclose(
                    test_doc["lattice_abc"], cart2abc(test_doc["lattice_cart"])
                ),
                msg="Conversion cart2abc failed.",
            )
            self.assertTrue(
                np.allclose(
                    cart2abc(test_doc["lattice_cart"]),
                    cart2abc(abc2cart(test_doc["lattice_abc"])),
                ),
                msg="Conversion abc2cart failed.",
            )
            self.assertAlmostEqual(
                test_doc["cell_volume"],
                cart2volume(test_doc["lattice_cart"]),
                msg="Failed to calculate volume from lattice vectors.",
                places=5,
            )
            self.assertIsInstance(
                test_doc["lattice_abc"], list, msg="Failed abc numpy cast to list"
            )
            self.assertIsInstance(
                test_doc["lattice_cart"],
                list,
                msg="Failed cartesian numpy cast to list",
            )
            cart_pos = frac2cart(test_doc["lattice_cart"], test_doc["positions_frac"])
            back2frac = cart2frac(test_doc["lattice_cart"], cart_pos)
            np.testing.assert_array_almost_equal(back2frac, test_doc["positions_frac"])
        except AssertionError:
            print("cart:", test_doc["lattice_cart"], abc2cart(test_doc["lattice_abc"]))
            print("abc:", test_doc["lattice_abc"], cart2abc(test_doc["lattice_cart"]))
            print(
                "volume:",
                test_doc["cell_volume"],
                cart2volume(test_doc["lattice_cart"]),
            )
            raise AssertionError

    def test_cart2abc_star(self):
        castep_fname = REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep"
        self.assertTrue(os.path.isfile(castep_fname))
        test_doc, success = castep2dict(castep_fname, db=True, verbosity=VERBOSITY)
        self.assertTrue(success)
        self.assertTrue(
            np.allclose(
                real2recip(test_doc["lattice_cart"]),
                2 * np.pi * np.asarray(cart2abcstar(test_doc["lattice_cart"])),
            ),
            msg="Conversion cart2abc failed.",
        )

    def test_frac2cart(self):
        lattice_cart = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        positions_frac = [0.5, 0.5, 0.5]
        self.assertEqual(frac2cart(lattice_cart, positions_frac), [1, 1, 1])

        lattice_cart = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        positions_frac = [[0.5, 0.5, 0.5]]
        self.assertEqual(frac2cart(lattice_cart, positions_frac), [[1, 1, 1]])

        lattice_cart = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        positions_frac = [[1, 1, 1], [0.5, 0.5, 0.5]]
        self.assertEqual(
            frac2cart(lattice_cart, positions_frac), [[2, 2, 2], [1, 1, 1]]
        )

    def test_cart2frac(self):
        lattice_cart = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        positions_abs = [1, 1, 1]
        self.assertEqual(cart2frac(lattice_cart, positions_abs), [0.5, 0.5, 0.5])

        lattice_cart = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        positions_abs = [[1, 1, 1]]
        self.assertEqual(cart2frac(lattice_cart, positions_abs), [[0.5, 0.5, 0.5]])

        lattice_cart = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        positions_abs = [[2, 2, 2], [1, 1, 1]]
        self.assertEqual(
            cart2frac(lattice_cart, positions_abs), [[1, 1, 1], [0.5, 0.5, 0.5]]
        )

    def test_conversion_transitivity(self):
        """ Test that cart2frac(frac2cart(A)) == A. """
        castep_fname = REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep"
        test_doc, s = castep2dict(castep_fname, db=True, verbosity=VERBOSITY)
        lattice_cart = test_doc["lattice_cart"]
        positions_frac = test_doc["positions_frac"]
        np.testing.assert_almost_equal(
            cart2frac(lattice_cart, frac2cart(lattice_cart, positions_frac)),
            positions_frac,
            decimal=10,
        )

    def test_random_noise(self):
        """ Test that zero amplitude random noise doesn't do anything, but any other
        amplitude does.
        """
        from matador.utils.cell_utils import add_noise

        castep_fname = REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep"
        test_doc, s = castep2dict(castep_fname, db=True, verbosity=VERBOSITY)

        init_positions_frac = test_doc["positions_frac"]
        test_doc = add_noise(test_doc, amplitude=0)
        np.testing.assert_almost_equal(
            init_positions_frac, test_doc["positions_frac"], decimal=10
        )

        test_doc = add_noise(test_doc, amplitude=0.1)
        np.testing.assert_almost_equal(
            init_positions_frac, test_doc["positions_frac"], decimal=1
        )
        diff = np.sqrt(
            np.sum(
                (
                    np.asarray(init_positions_frac)
                    - np.asarray(test_doc["positions_frac"])
                )
                ** 2,
                axis=-1,
            )
        )
        self.assertTrue(np.max(diff) <= 0.1)
        self.assertTrue(np.max(diff) >= 0)

    def test_recip2real(self):
        real_lattice = [
            [5.5902240, 0, 0],
            [3.7563195, 4.1401290, 0],
            [-2.9800295, -1.3200288, 8.5321695],
        ]
        recip_lattice = real2recip(real_lattice)
        np.testing.assert_array_almost_equal(
            np.asarray(recip_lattice),
            np.asarray(
                [
                    [1.1239595, -1.0197632, 0.2347956],
                    [0.0, 1.5176303, 0.2347956],
                    [0, 0, 0.7364112],
                ]
            ),
        )

        real_lattice = [
            [6.0235150, 0, 0],
            [0.0, 5.6096010, 0],
            [-5.0202472, 0, 10.0218337],
        ]
        recip_lattice = real2recip(real_lattice)
        np.testing.assert_array_almost_equal(
            np.asarray(recip_lattice),
            np.asarray(
                [[1.0431094, 0, 0.5225256], [0, 1.1200770, 0], [0, 0, 0.6269494]]
            ),
        )

    def test_calc_mp_spacing(self):
        from matador.utils.cell_utils import calc_mp_spacing

        real_lattice = [
            [6.0235150, 0, 0],
            [0.0, 5.6096010, 0],
            [-5.0202472, 0, 10.0218337],
        ]
        mp_grid = [4, 4, 2]
        spacing = calc_mp_spacing(real_lattice, mp_grid, prec=3)
        self.assertEqual(spacing, 0.05)
        spacing = calc_mp_spacing(real_lattice, mp_grid, prec=2)
        self.assertAlmostEqual(spacing, 0.05)
        spacing = calc_mp_spacing(real_lattice, mp_grid, prec=5)
        self.assertAlmostEqual(spacing, 0.05, places=3)

    def test_calc_mp_grid(self):
        from matador.utils.cell_utils import calc_mp_grid

        real_lattice = [
            [6.0235150, 0, 0],
            [0.0, 5.6096010, 0],
            [-5.0202472, 0, 10.0218337],
        ]
        spacing = 0.05
        self.assertEqual(calc_mp_grid(real_lattice, spacing), [4, 4, 2])

    def test_shift_grid(self):
        from matador.utils.cell_utils import (
            calc_mp_grid,
            shift_to_include_gamma,
            abc2cart,
        )

        lattice_abc = [[5.57068, 10.222092, 10.222039], [90.0, 90.0, 90.0]]
        lattice_cart = abc2cart(lattice_abc)
        spacing = 0.1
        mp_grid = calc_mp_grid(lattice_cart, spacing)
        self.assertEqual(mp_grid, [2, 1, 1])
        self.assertEqual(shift_to_include_gamma(mp_grid), [0.25, 0, 0])

        lattice_abc = [[5.57068, 10.222092, 10.222039], [90.0, 90.0, 90.0]]
        lattice_cart = abc2cart(lattice_abc)
        spacing = 0.05
        mp_grid = calc_mp_grid(lattice_cart, spacing)
        self.assertEqual(mp_grid, [4, 2, 2])
        self.assertEqual(shift_to_include_gamma(mp_grid), [0.125, 0.25, 0.25])

    def test_spg_label(self):
        from matador.utils.cell_utils import get_space_group_label_latex

        test_cases = ["P-63m", "Fm-3m", "I4/mmm", "P4_2/mmc"]
        results = ["$P\\bar{6}3m$", "$Fm\\bar{3}m$", "$I4/mmm$", "$P4_2/mmc$"]
        for res, test in zip(results, test_cases):
            self.assertEqual(res, get_space_group_label_latex(test))


class SymmetriesAndSupercellsTest(unittest.TestCase):
    """ Tests cell util functions. """

    def test_supercell_creator(self):
        castep_fname = REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep"
        self.assertTrue(os.path.isfile(castep_fname))
        num_tests = 3
        test_doc, s = castep2dict(castep_fname, db=True, verbosity=VERBOSITY)
        _iter = 0
        while _iter < num_tests:
            extension = np.random.randint(low=1, high=5, size=(3)).tolist()
            if extension == [1, 1, 1]:
                extension[np.random.randint(low=0, high=2)] += 1
            num_images = np.prod(extension)

            standardize = bool(_iter % 2)
            symmetric = bool(_iter % 2)

            supercell = create_simple_supercell(
                test_doc, tuple(extension), standardize=standardize, symmetric=symmetric
            )
            self.assertEqual(supercell["num_atoms"], num_images * test_doc["num_atoms"])
            self.assertAlmostEqual(
                supercell["cell_volume"], num_images * test_doc["cell_volume"], places=3
            )
            self.assertEqual(
                len(supercell["positions_frac"]),
                num_images * len(test_doc["positions_frac"]),
            )
            for i in range(3):
                if not standardize:
                    np.testing.assert_array_equal(
                        np.asarray(supercell["lattice_cart"][i]),
                        extension[i] * np.asarray(test_doc["lattice_cart"][i]),
                    )
            self.assertLess(pdf_sim_dist(test_doc, supercell), 1e-3)
            _iter += 1

        # test error for 1x1x1
        with self.assertRaises(RuntimeError):
            supercell = create_simple_supercell(test_doc, (1, 1, 1))

    @unittest.skipIf(
        not IMPORTED_SEEKPATH, "Seekpath package not found in this distribution"
    )
    def test_kpt_path(self):

        cell, s = castep2dict(REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep")
        std_cell, path, seekpath_results = get_seekpath_kpoint_path(
            cell, spacing=0.01, debug=False
        )
        self.assertEqual(539, len(path))

        self.assertLess(pdf_sim_dist(cell, std_cell), 0.05)

        import glob
        from os import remove
        from matador.utils.cell_utils import frac2cart

        fnames = glob.glob(REAL_PATH + "data/bs_test/*.res")
        spacing = 0.01
        for fname in fnames:
            doc, s = res2dict(fname, db=False)
            doc["cell_volume"] = cart2volume(doc["lattice_cart"])

            std_doc, path, seekpath_results = get_seekpath_kpoint_path(
                doc, spacing=spacing, debug=False
            )
            seekpath_results_path = get_path(doc2spg(doc))

            rel_path = seekpath_results["explicit_kpoints_rel"]
            abs_path = seekpath_results["explicit_kpoints_abs"]

            cart_kpts = np.asarray(frac2cart(real2recip(std_doc["lattice_cart"]), path))
            diffs = np.zeros((len(cart_kpts[:-1])))
            np.testing.assert_array_almost_equal(cart_kpts, abs_path)
            np.testing.assert_array_almost_equal(path, rel_path)
            for ind, kpt in enumerate(cart_kpts[:-1]):
                diffs[ind] = np.sqrt(np.sum((kpt - cart_kpts[ind + 1]) ** 2))
            self.assertLess(
                len(np.where(diffs > 1.1 * spacing)[0]),
                len(seekpath_results["explicit_segments"]),
            )

            if "flrys4-1x109" in fname:
                bs, s = bands2dict(fname.replace(".res", ".bands"))
                np.testing.assert_array_almost_equal(bs["kpoint_path"], rel_path)
                np.testing.assert_array_almost_equal(
                    bs["lattice_cart"], std_doc["lattice_cart"]
                )
            self.assertLess(
                len(np.where(diffs > 1.1 * spacing)[0]),
                len(seekpath_results["explicit_segments"]),
            )

            cell_path = fname.replace(".res", ".cell")
            doc2cell(std_doc, cell_path)
            new_doc, s = cell2dict(cell_path, lattice=True, positions=True, db=False)
            assert "positions_frac" in new_doc
            remove(cell_path)
            seekpath_new_results = get_path(doc2spg(new_doc))
            self.assertEqual(
                seekpath_new_results["bravais_lattice_extended"],
                seekpath_results_path["bravais_lattice_extended"],
            )

            dist = pdf_sim_dist(doc, std_doc)
            self.assertLess(dist, 0.01)
            dist = pdf_sim_dist(doc, new_doc)
            self.assertLess(dist, 0.01)

    def test_spg_standardize(self):
        from matador.utils.cell_utils import standardize_doc_cell
        from matador.scrapers import cif2dict
        import glob

        doc, s = castep2dict(REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep")
        std_doc = standardize_doc_cell(doc)
        dist = pdf_sim_dist(doc, std_doc)
        self.assertLess(dist, 0.01)

        fnames = glob.glob(REAL_PATH + "data/bs_test/*.res")
        for fname in fnames:
            doc, s = res2dict(fname, db=False)
            doc["cell_volume"] = cart2volume(doc["lattice_cart"])
            std_doc = standardize_doc_cell(doc)
            dist = pdf_sim_dist(doc, std_doc)
            self.assertLess(dist, 0.01)

        doc = Crystal(
            castep2dict(REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep")[0]
        )
        std_doc = standardize_doc_cell(doc)
        dist = pdf_sim_dist(doc, std_doc)
        self.assertLess(dist, 0.01)

        doc = Crystal(cif2dict(REAL_PATH + "data/cif_files/AgBiI.cif")[0])
        with self.assertRaises(RuntimeError):
            std_doc = standardize_doc_cell(doc)


def pdf_sim_dist(doc_test, doc_supercell):
    doc_test["text_id"] = ["test", "cell"]
    doc_supercell["text_id"] = ["super", "cell"]
    pdf_test = PDF(doc_test, low_mem=True)
    pdf_supercell = PDF(doc_supercell, low_mem=True)
    overlap = PDFOverlap(pdf_test, pdf_supercell)
    return overlap.similarity_distance


if __name__ == "__main__":
    unittest.main()



""" This file contains a Python list of all CASTEP parameters,
automatically generated with file_utils.scrape_castep_params().
"""

CASTEP_VERSION = 18.1

CASTEP_CELL_KEYWORDS = [
    "atomic_init",
    "bs_kpoints_list",
    "bs_kpoints_mp_grid",
    "bs_kpoints_mp_offset",
    "bs_kpoints_mp_spacing",
    "bs_kpoints_path",
    "bs_kpoints_path_spacing",
    "bs_kpoint_mp_grid",
    "bs_kpoint_mp_offset",
    "bs_kpoint_mp_spacing",
    "bs_kpoint_path_spacing",
    "cell_noise",
    "chemical_potential",
    "elnes_kpoint_mp_grid",
    "elnes_kpoint_mp_offset",
    "elnes_kpoint_mp_spacing",
    "fix_all_cell",
    "fix_all_ions",
    "fix_com",
    "fix_vol",
    "hubbard_alpha",
    "ionic_constraints",
    "jcoupling_site",
    "kpoints_mp_grid",
    "kpoints_mp_offset",
    "kpoints_mp_spacing",
    "kpoint_list",
    "kpoint_mp_grid",
    "kpoint_mp_offset",
    "kpoint_mp_spacing",
    "magres_kpoint_mp_grid",
    "magres_kpoint_mp_offset",
    "magres_kpoint_mp_spacing",
    "magres_kpoint_path",
    "magres_kpoint_path_spacing",
    "nonlinear_constraints",
    "optics_kpoints_mp_grid",
    "optics_kpoints_mp_offset",
    "optics_kpoints_mp_spacing",
    "optics_kpoint_list",
    "optics_kpoint_mp_grid",
    "optics_kpoint_mp_offset",
    "optics_kpoint_mp_spacing",
    "phonon_fine_kpoint_mp_grid",
    "phonon_fine_kpoint_mp_offset",
    "phonon_fine_kpoint_mp_spacing",
    "phonon_fine_kpoint_path",
    "phonon_fine_kpoint_path_spacing",
    "phonon_gamma_directions",
    "phonon_kpoints_list",
    "phonon_kpoints_path",
    "phonon_kpoints_path_spacing",
    "phonon_kpoint_mp_grid",
    "phonon_kpoint_mp_offset",
    "phonon_kpoint_mp_spacing",
    "phonon_kpoint_path",
    "phonon_kpoint_path_spacing",
    "phonon_supercell_matrix",
    "positions_abs_intermediate",
    "positions_abs_product",
    "positions_frac_intermediate",
    "positions_frac_product",
    "positions_noise",
    "quantisation_axis",
    "quantization_axis",
    "sedc_custom_params",
    "snap_to_symmetry",
    "species_gamma",
    "species_lcao_states",
    "species_q",
    "spectral_kpoints_list",
    "spectral_kpoints_mp_grid",
    "spectral_kpoints_mp_offset",
    "spectral_kpoints_mp_spacing",
    "spectral_kpoints_path",
    "spectral_kpoints_path_spacing",
    "spectral_kpoint_mp_grid",
    "spectral_kpoint_mp_offset",
    "spectral_kpoint_mp_spacing",
    "spectral_kpoint_path",
    "spectral_kpoint_path_spacing",
    "supercell_kpoints_mp_grid",
    "supercell_kpoints_mp_offset",
    "supercell_kpoints_mp_spacing",
    "supercell_kpoint_list",
    "supercell_kpoint_mp_grid",
    "supercell_kpoint_mp_offset",
    "supercell_kpoint_mp_spacing",
    "symmetry_generate",
    "symmetry_tol",
    "unit_cell",
]
CASTEP_CELL_BLOCKS = [
    "bs_kpoint_list",
    "bs_kpoint_path",
    "cell_constraints",
    "elnes_kpoint_list",
    "external_efield",
    "external_pressure",
    "hubbard_u",
    "help",
    "ionic_velocities",
    "kpoints_list",
    "lattice_abc",
    "lattice_cart",
    "magres_kpoint_list",
    "optics_kpoints_list",
    "phonon_fine_kpoint_list",
    "phonon_kpoint_list",
    "positions_abs",
    "positions_frac",
    "species_mass",
    "species_pot",
    "spectral_kpoint_list",
    "supercell_kpoints_list",
    "supercell_matrix",
    "symmetry_ops",
]
CASTEP_PARAMS = [
    "backup_interval",
    "basis_de_dloge",
    "basis_precision",
    "born_charge_sum_rule",
    "bs_eigenvalue_tol",
    "bs_max_cg_steps",
    "bs_max_iter",
    "bs_nbands",
    "bs_nextra_bands",
    "bs_perc_extra_bands",
    "bs_re_est_k_scrn",
    "bs_write_eigenvalues",
    "bs_xc_definition",
    "bs_xc_functional",
    "calculate_born_charges",
    "calculate_densdiff",
    "calculate_elf",
    "calculate_hirshfeld",
    "calculate_raman",
    "calculate_stress",
    "calc_full_ex_pot",
    "calc_molecular_dipole",
    "charge",
    "charge_unit",
    "checkpoint",
    "cml_filename",
    "cml_output",
    "comment",
    "continuation",
    "cut_off_energy",
    "data_distribution",
    "devel_code",
    "dipole_correction",
    "dipole_dir",
    "dipole_unit",
    "efermi_tol",
    "efield_calc_ion_permittivity",
    "efield_convergence_win",
    "efield_dfpt_method",
    "efield_energy_tol",
    "efield_freq_spacing",
    "efield_ignore_molec_modes",
    "efield_max_cg_steps",
    "efield_max_cycles",
    "efield_oscillator_q",
    "efield_unit",
    "electronic_minimizer",
    "elec_convergence_win",
    "elec_dump_file",
    "elec_eigenvalue_tol",
    "elec_energy_tol",
    "elec_force_tol",
    "elec_method",
    "elec_restore_file",
    "elec_temp",
    "elnes_eigenvalue_tol",
    "elnes_nbands",
    "elnes_nextra_bands",
    "elnes_perc_extra_bands",
    "elnes_xc_definition",
    "elnes_xc_functional",
    "energy_unit",
    "entropy_unit",
    "exchange_reflect_kpts",
    "excited_state_scissors",
    "fft_max_prime_factor",
    "fine_cut_off_energy",
    "fine_gmax",
    "fine_grid_scale",
    "finite_basis_corr",
    "finite_basis_npoints",
    "finite_basis_spacing",
    "fixed_npw",
    "fix_occupancy",
    "force_constant_unit",
    "force_unit",
    "frequency_unit",
    "ga_bulk_slice",
    "ga_fixed_n",
    "ga_max_gens",
    "ga_mutate_amp",
    "ga_mutate_rate",
    "ga_pop_size",
    "geom_convergence_win",
    "geom_disp_tol",
    "geom_energy_tol",
    "geom_force_tol",
    "geom_frequency_est",
    "geom_lbfgs_max_updates",
    "geom_linmin_tol",
    "geom_max_iter",
    "geom_method",
    "geom_modulus_est",
    "geom_spin_fix",
    "geom_stress_tol",
    "geom_tpsd_init_stepsize",
    "geom_tpsd_iterchange",
    "geom_use_linmin",
    "grid_scale",
    "impose_trs",
    "inv_length_unit",
    "iprint",
    "ir_intensity_unit",
    "k_scrn_averaging_scheme",
    "k_scrn_den_function",
    "length_unit",
    "magres_convergence_win",
    "magres_conv_tol",
    "magres_jcoupling_task",
    "magres_max_cg_steps",
    "magres_max_sc_cycles",
    "magres_method",
    "magres_task",
    "magres_write_response",
    "magres_xc_definition",
    "magres_xc_functional",
    "mass_unit",
    "max_cg_steps",
    "max_diis_steps",
    "max_scf_cycles",
    "max_sd_steps",
    "md_barostat",
    "md_cell_t",
    "md_damping_reset",
    "md_damping_scheme",
    "md_delta_t",
    "md_elec_convergence_win",
    "md_elec_eigenvalue_tol",
    "md_elec_energy_tol",
    "md_elec_force_tol",
    "md_ensemble",
    "md_eqm_cell_t",
    "md_eqm_ion_t",
    "md_eqm_method",
    "md_eqm_t",
    "md_extrap",
    "md_extrap_fit",
    "md_hug_compression",
    "md_hug_dir",
    "md_hug_method",
    "md_hug_t",
    "md_ion_t",
    "md_langevin_t",
    "md_nhc_length",
    "md_nose_t",
    "md_num_beads",
    "md_num_iter",
    "md_opt_damped_delta_t",
    "md_pathint_init",
    "md_pathint_num_stages",
    "md_pathint_staging",
    "md_sample_iter",
    "md_temperature",
    "md_thermostat",
    "md_use_pathint",
    "md_use_plumed",
    "md_xlbomd",
    "md_xlbomd_history",
    "message_size",
    "metals_method",
    "mixing_scheme",
    "mix_charge_amp",
    "mix_charge_gmax",
    "mix_cut_off_energy",
    "mix_history_length",
    "mix_metric_q",
    "mix_spin_amp",
    "mix_spin_gmax",
    "nbands",
    "ndown",
    "nelectrons",
    "nextra_bands",
    "nlxc_calc_full_ex_pot",
    "nlxc_div_corr_npts_step",
    "nlxc_div_corr_on",
    "nlxc_div_corr_s_width",
    "nlxc_div_corr_tol",
    "nlxc_exchange_fraction",
    "nlxc_exchange_reflect_kpts",
    "nlxc_exchange_screening",
    "nlxc_impose_trs",
    "nlxc_k_scrn_averaging_scheme",
    "nlxc_k_scrn_den_function",
    "nlxc_page_ex_pot",
    "nlxc_ppd_integral",
    "nlxc_ppd_size_x",
    "nlxc_ppd_size_y",
    "nlxc_ppd_size_z",
    "nlxc_re_est_k_scrn",
    "nspins",
    "num_backup_iter",
    "num_dump_cycles",
    "num_farms",
    "num_occ_cycles",
    "num_proc_in_smp",
    "num_proc_in_smp_fine",
    "nup",
    "optics_nbands",
    "optics_nextra_bands",
    "optics_perc_extra_bands",
    "optics_xc_definition",
    "optics_xc_functional",
    "opt_strategy",
    "opt_strategy_bias",
    "page_ex_pot",
    "page_wvfns",
    "pdos_calculate_weights",
    "perc_extra_bands",
    "phonon_calculate_dos",
    "phonon_calc_lo_to_splitting",
    "phonon_const_basis",
    "phonon_convergence_win",
    "phonon_dfpt_method",
    "phonon_dos_limit",
    "phonon_dos_spacing",
    "phonon_energy_tol",
    "phonon_fine_cutoff_method",
    "phonon_fine_method",
    "phonon_finite_disp",
    "phonon_force_constant_cutoff",
    "phonon_force_constant_cut_scale",
    "phonon_force_constant_ellipsoid",
    "phonon_max_cg_steps",
    "phonon_max_cycles",
    "phonon_method",
    "phonon_preconditioner",
    "phonon_sum_rule",
    "phonon_sum_rule_method",
    "phonon_use_kpoint_symmetry",
    "phonon_write_dynamical",
    "phonon_write_force_constants",
    "popn_bond_cutoff",
    "popn_calculate",
    "popn_write",
    "ppd_integral",
    "ppd_size_x",
    "ppd_size_y",
    "ppd_size_z",
    "pressure_unit",
    "print_clock",
    "print_memory_usage",
    "pspot_beta_phi_type",
    "pspot_nonlocal_type",
    "raman_method",
    "raman_range_high",
    "raman_range_low",
    "rand_seed",
    "relativistic_treatment",
    "reuse",
    "re_est_k_scrn",
    "run_time",
    "secondd_method",
    "sedc_apply",
    "sedc_d_g06",
    "sedc_d_jchs",
    "sedc_d_ts",
    "sedc_lambda_obs",
    "sedc_n_obs",
    "sedc_s6_g06",
    "sedc_s6_jchs",
    "sedc_scheme",
    "sedc_sr_jchs",
    "sedc_sr_ts",
    "smearing_scheme",
    "smearing_width",
    "spectral_eigenvalue_tol",
    "spectral_max_iter",
    "spectral_max_steps_per_iter",
    "spectral_nbands",
    "spectral_nextra_bands",
    "spectral_perc_extra_bands",
    "spectral_re_est_k_scrn",
    "spectral_task",
    "spectral_theory",
    "spectral_write_eigenvalues",
    "spectral_xc_definition",
    "spectral_xc_functional",
    "spin",
    "spin_fix",
    "spin_orbit_coupling",
    "spin_polarised",
    "spin_polarized",
    "spin_treatment",
    "spin_unit",
    "stop",
    "task",
    "tddft_approximation",
    "tddft_convergence_win",
    "tddft_eigenvalue_method",
    "tddft_eigenvalue_tol",
    "tddft_max_iter",
    "tddft_method",
    "tddft_nextra_states",
    "tddft_num_states",
    "tddft_position_method",
    "tddft_selected_state",
    "tddft_xc_definition",
    "tddft_xc_functional",
    "thermo_calculate_helmholtz",
    "thermo_t_npoints",
    "thermo_t_spacing",
    "thermo_t_start",
    "thermo_t_stop",
    "time_unit",
    "tssearch_cg_max_iter",
    "tssearch_disp_tol",
    "tssearch_energy_tol",
    "tssearch_force_tol",
    "tssearch_lstqst_protocol",
    "tssearch_max_path_points",
    "tssearch_method",
    "tssearch_qst_max_iter",
    "velocity_unit",
    "verbosity",
    "volume_unit",
    "wannier_ion_cmoments",
    "wannier_ion_cut",
    "wannier_ion_cut_fraction",
    "wannier_ion_cut_tol",
    "wannier_ion_moments",
    "wannier_ion_rmax",
    "wannier_max_sd_steps",
    "wannier_min_algor",
    "wannier_print_cube",
    "wannier_restart",
    "wannier_sd_step",
    "wannier_spread_tol",
    "wannier_spread_type",
    "write_bands",
    "write_bib",
    "write_cell_structure",
    "write_checkpoint",
    "write_cif_structure",
    "write_cst_esp",
    "write_formatted_density",
    "write_formatted_elf",
    "write_formatted_potential",
    "write_geom",
    "write_md",
    "write_none",
    "write_orbitals",
    "write_otfg",
    "xc_definition",
    "xc_functional",
    "xc_vxc_deriv_epsilon",
]



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the AtomicSwapper class which
takes a Query object and a desired "swap", e.g. swap all group [I]
elements to Li, ready to be re-relaxed.

"""


__all__ = ['AtomicSwapper']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.swaps.swaps import AtomicSwapper



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements filtering based on Fingerprint objects,
although only PDF has been implemented so far.

"""

import copy
from collections import defaultdict
import numpy as np
from matador.fingerprints.pdf import PDF, PDFFactory
from matador.utils.cursor_utils import get_guess_doc_provenance


def get_uniq_cursor(cursor, sim_tol=0.1, energy_tol=1e-2,
                    enforce_same_stoich=True, fingerprint=PDF,
                    hierarchy_order=None, hierarchy_values=None,
                    debug=False, **fingerprint_calc_args):
    """ Uses fingerprint to filter cursor into unique structures to some
    tolerance sim_tol, additionally returning a dict of duplicates and the
    correlation matrix.

    The choice of which of the dulpicates is kept in the unique cursor is
    defined by the "hierarchy". By default, this will guess the provenance
    of a document and prefer structures from "primary sources", i.e.
    ICSD -> OQMD -> Materials Project -> SWAPS -> AIRSS -> GA. A custom hiearchy
    can be provided through `hierarchy_order`, which must be accompanied by a list
    of values per structure to check against that hierarchy.

    Parameters:
        cursor (list) : matador cursor to be filtered

    Keyword Arguments:
        fingerprint (Fingerprint): fingerprint object type to compare
            (DEFAULT: PDF)
        sim_tol (float/bool): tolerance in similarity distance for
            duplicates (if True, default value of 0.1 used)
        energy_tol (float): compare only structures within a certain
            energy tolerance (1e20 if enforce_same_stoich is False)
        enforce_same_stoich (bool): compare only structures of the same
            stoichiometry
        debug (bool): print timings and list similarities
        fingerprint_calc_args (dict): kwargs to pass to fingerprint

    Returns:
        distinct_set (list): ordered list indices of unique documents
        dupe_dict (dict): a dict with keys from distinct_set, listing duplicates
        fingerprint_list (list): a list of <Fingerprint> objects
        sim_mat (np.ndarray): the correlation matrix of pair similarity distances

    """

    if isinstance(sim_tol, bool):
        sim_tol = 0.1

    if not cursor:
        raise RuntimeError("No structures provided to compare.")

    fingerprint_list = []
    if not enforce_same_stoich:
        energy_tol = 1e20
    print('Calculating fingerprints...')

    fingerprint_list = [None for doc in cursor]
    required_inds = set()

    # scipy sparse matrices dont seem to allow non-zero default values, so we'll use a defaultdict
    sim_mat = defaultdict(lambda: 1e10)
    print('Assessing similarities...')
    for i in range(len(fingerprint_list)):
        for j in range(i+1, len(fingerprint_list)):
            # are we checking stoichiometries, if so, ensure they're the same
            if (enforce_same_stoich is False or
                    (sorted(cursor[j]['stoichiometry']) == sorted(cursor[i]['stoichiometry']) and
                     np.abs(cursor[j].get('enthalpy_per_atom', 0) - cursor[i].get('enthalpy_per_atom', 0)) < energy_tol)):
                # need to set both to zero so we can iterate over the dict later
                sim_mat[i, j] = None
                sim_mat[j, i] = None
                required_inds.add(i)
                required_inds.add(j)

    factory = PDFFactory(cursor, required_inds=list(required_inds), **fingerprint_calc_args)

    for i, j in sim_mat:
        if sim_mat[i, j] is None:
            sim = cursor[i][factory.default_key].get_sim_distance(cursor[j][factory.default_key])
            sim_mat[i, j] = sim
            sim_mat[j, i] = sim

    distinct_set = set()
    dupe_set = set()
    dupe_dict = dict()
    for i in range(len(cursor)):
        distinct_set.add(i)
        dupe_dict[i] = set()

    # loop over the similarity matrix and construct the set of "unique" structures
    # and a dictionary containing their duplicates
    for i, j in sim_mat:
        if sim_mat[i, j] <= sim_tol:
            if i not in dupe_set:
                if j in distinct_set:
                    distinct_set.remove(j)
                    del dupe_dict[j]
                dupe_set.add(j)
                dupe_dict[i].add(j)

    total_dupes = len(set(list(dupe_dict.keys()) + [item for key in dupe_dict for item in dupe_dict[key]]))
    if len(cursor) != total_dupes:
        raise RuntimeError("Something went wrong: dupe dict had wrong size {} compared to cursor {}!\nFull output: {}"
                           .format(total_dupes, len(cursor), dupe_dict))

    if hierarchy_order is None:
        hierarchy_order = ['ICSD', 'DOI', 'OQMD', 'MP', 'PF', 'SWAPS', 'AIRSS', 'GA']
    if hierarchy_values is None:
        hierarchy_values = [get_guess_doc_provenance(doc['source']) for doc in cursor]

    print('Applying hierarchy of structures with order: {}'.format(hierarchy_order))
    dupe_dict = _enforce_hierarchy(dupe_dict, hierarchy_values, hierarchy_order)

    all_structures = set(list(dupe_dict.keys()) + [item for key in dupe_dict for item in dupe_dict[key]])
    if len(cursor) != len(all_structures):
        raise RuntimeError("Something went wrong: dupe dict had wrong size {} compared to cursor {}!\nDifference: {}"
                           .format(len(all_structures),
                                   len(cursor),
                                   all_structures.symmetric_difference({i for i in range(len(cursor))})))

    print('Done!')
    return sorted(list(dupe_dict.keys())), dupe_dict, fingerprint_list, sim_mat


def _enforce_hierarchy(dupe_dict, values, hierarchy):
    """ Enforce a general hierarchy of which structures to keep, based
    on the list of values and their importance.

    Parameters:
        dupe_dict (dict): the dictionary keyed by the index of unique structures
            that holds lists of duplicates for that structure.
        values (list): the list of values for each structure on which to enforce
            the hierarchy.
        hierarchy (list): the order in which to consider the values, e.g.
            `['ICSD', 'OQMD']` will promote ICSD structures over OQMD.

    Returns:
        dict: the reshuffled dictionary of duplicates.

    """
    max_val = max(list(dupe_dict.keys()) + [val for t in dupe_dict.values() for val in t])

    if len(values) - 1 != max_val:
        raise RuntimeError("Number of hierarchy values does not much number of items: {} vs {}"
                           .format(len(values)-1, max_val))

    new_dupe_dict = copy.deepcopy(dupe_dict)

    swapped = []
    for i in new_dupe_dict:
        if not list(new_dupe_dict[i]):
            continue
        for value in hierarchy:
            found = False
            for k in [i] + list(new_dupe_dict[i]):
                if values[k] == value:
                    swapped.append((i, k))
                    found = True
                    break
            if found:
                break

    for i, k in swapped:
        if i != k:
            if k in new_dupe_dict:
                new_dupe_dict[k].update([ind for ind in new_dupe_dict[i] if ind != k] + [i])
            else:
                new_dupe_dict[k] = set([ind for ind in new_dupe_dict[i] if ind != k] + [i])
            del new_dupe_dict[i]

    return new_dupe_dict



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the loading custom settings for plotting and
for database collections.

"""


import os
from pathlib import Path
from traceback import print_exc
from matador.utils.print_utils import print_warning
from matador.config.settings import Settings
from matador.config.quickstart import quickstart_settings

SETTINGS = Settings()

DEFAULT_SETTINGS = {
    "mongo": {
        "db": "crystals",
        "host": "localhost",
        "port": 27017,
        "default_collection": "repo",
    },
    "plotting": {"default_style": "matador", "element_colours": "vesta_elements.ini"},
    "run3": {
        "scratch_prefix": ".",
        "castep_executable": "castep",
        "optados_executable": "optados",
    },
}

FILE_PATHS = {
    "mongo": ["default_collection_file_path", "scratch_prefix"],
    "plotting": ["element_colours"],
}


def load_custom_settings(
    config_fname=None, quiet=False, debug=False, no_quickstart=True
):
    """ Load mongodb settings dict from file given by fname, or from
    defaults. Hierarchy of filenames to check:

        1. .matadorrc
        2. ~/.matadorrc
        3. ~/.config/matadorrc

    Keyword Arguments:
        fname (str): filename of custom settings file.
        quiet (bool): print nothing on loading.
        no_quickstart (bool): for testing purposes, override any questions.
        debug (bool): print settings on loading.

    """
    if SETTINGS.set:
        return SETTINGS

    import yaml

    if config_fname is None:
        trial_user_fnames = [
            Path(".matadorrc"),
            Path(os.path.expanduser("~/.matadorrc")),
            Path(os.path.expanduser("~/.config/matadorrc")),
            Path(__file__).parent.joinpath("matadorrc.yml"),
        ]

        # use first file that exists in hierarchy
        for fname in trial_user_fnames:
            if os.path.isfile(fname):
                config_fname = str(fname)
                break
        else:
            # otherwise offer to create one, or use default
            if not no_quickstart:
                config_fname = quickstart_settings()
            if config_fname is None:
                # otherwise load default
                print("No config provided: loading perhaps unsuitable defaults instead.")
                config_fname = Path(__file__).parent.joinpath("matadorrc.yml"),

        if not quiet:
            print("Loading settings from {}".format(config_fname))

    custom_settings = {}
    if os.path.isfile(config_fname):
        try:
            with open(config_fname, "r") as f:
                custom_settings = yaml.safe_load(f)
                if custom_settings is None:
                    custom_settings = {}
        except Exception:
            print_exc()
            raise SystemExit(
                "Failed to read custom settings file {} as YAML".format(config_fname)
            )
    else:
        print_warning(
            "Could not find {}, loading default settings...".format(config_fname)
        )

    settings = {}
    settings.update(DEFAULT_SETTINGS)
    settings.update(custom_settings)

    # check absolute/relative paths
    for key in FILE_PATHS:
        for setting in FILE_PATHS[key]:
            if settings.get(key) is not None:
                if settings[key].get(setting) is not None:
                    settings[key][setting] = str(
                        Path(os.path.expanduser(settings[key][setting])).resolve()
                    )

    if debug:
        import json
        print(json.dumps(settings, indent=2))

    set_settings(settings, override=False)

    return SETTINGS


def set_settings(settings, override=True):
    """ Updates the global settings with the dictionary
    provided.

    Parameters:
        settings (dict): the key-value pairs to use.

    Keyword arguments:
        override (bool): whether or not to override
            previous settings, or just update.

    """

    if SETTINGS.set and not override:
        return

    keys = list(SETTINGS.settings.keys())
    for key in keys:
        if key not in settings:
            SETTINGS.settings.pop(key)

    for key in settings:
        SETTINGS[key] = settings[key]

    SETTINGS.set = True



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the `TemperatureDependentHull` class
for assessing phase stability from finite temperature free energies.

"""

import copy
import numpy as np

from matador.hull.hull_ensemble import EnsembleHull
from matador.orm.spectral import VibrationalDOS


class TemperatureDependentHull(EnsembleHull):
    """ Leverages `EnsembleHull` to construct temperature dependent
    hulls from phonon calculations.

    """

    data_key = "temperature"
    energy_key = "free_energy_per_atom"

    def __init__(self, cursor, energy_key='enthalpy_per_atom', temperatures=None, **kwargs):

        self.temperatures = temperatures
        if temperatures is None:
            self.temperatures = np.linspace(0, 800, 21)

        _cursor = copy.deepcopy(cursor)

        # prepare the cursor by computing free energies
        # and store it in the format expected by EnsembleHull
        for ind, doc in enumerate(cursor):
            if not isinstance(doc, VibrationalDOS):
                _doc = VibrationalDOS(doc)
                temps, vib_free_energies = _doc.vibrational_free_energy(temperatures=temperatures)
                _cursor[ind][self.data_key] = {}
                _cursor[ind][self.data_key][self.energy_key] = np.ones_like(temperatures) * _cursor[ind][energy_key]
                _cursor[ind][self.data_key][self.energy_key] += vib_free_energies
                _cursor[ind][self.data_key]["temperatures"] = temperatures

        super().__init__(
            cursor=_cursor,
            data_key=self.data_key,
            energy_key=self.energy_key,
            chempot_energy_key=energy_key,
            parameter_key="temperatures",
            **kwargs
        )

    def plot_hull(self, **kwargs):
        """ Hull plot helper function. """
        from matador.plotting.hull_plotting import plot_temperature_hull
        return plot_temperature_hull(self, **kwargs)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot battery-relevant curves,
such as voltages and volume expansions.

"""

from typing import List, Optional, Dict, Union

import numpy as np
from matador.utils.chem_utils import get_formula_from_stoich
from matador.plotting.plotting import plotting_function, SAVE_EXTS
from matador.battery import VoltageProfile


__all__ = ['plot_voltage_curve', 'plot_volume_curve']


@plotting_function
def plot_voltage_curve(
    profiles: Union[List[VoltageProfile], VoltageProfile],
    ax=None,
    show: bool = False,
    labels: bool = False,
    savefig: Optional[str] = None,
    curve_labels: Optional[Union[str, List[str]]] = None,
    line_kwargs: Optional[Union[Dict, List[Dict]]] = None,
    expt: Optional[str] = None,
    expt_label: Optional[str] = None
):
    """ Plot voltage curve calculated for phase diagram.

    Parameters:
        profiles (list/VoltageProfile): list of/single voltage profile(s).

    Keyword arguments:
        ax (matplotlib.axes.Axes): an existing axis on which to plot.
        show (bool): whether to show plot in an X window.
        savefig (str): filename to use to save the plot.
        curve_labels (list): optional list of labels for the curves in
            the profiles list.
        line_kwargs (list or dict): parameters to pass to the curve plotter,
            if a list then the line kwargs will be passed to each line individually.
        expt (str): string to a filename of a csv Q, V to add to the plot.
        expt_label (str): label for any experimental profile passed to the plot.

    """
    import matplotlib.pyplot as plt

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    if not isinstance(profiles, list):
        profiles = [profiles]
    if curve_labels is not None and not isinstance(curve_labels, list):
        curve_labels = [curve_labels]
    if line_kwargs is not None and not isinstance(line_kwargs, list):
        line_kwargs = [line_kwargs]

    if curve_labels is not None and len(curve_labels) != len(profiles):
        raise RuntimeError(
            "Wrong number of labels passed for number of profiles: {} vs {}"
            .format(len(curve_labels), len(profiles))
        )

    if line_kwargs is not None and len(line_kwargs) != len(profiles):
        raise RuntimeError(
            "Wrong number of line kwargs passed for number of profiles: {} vs {}"
            .format(len(line_kwargs), len(profiles))
        )

    dft_label = None
    if expt is not None:
        expt_data = np.loadtxt(expt, delimiter=',')
        if expt_label:
            ax.plot(expt_data[:, 0], expt_data[:, 1], c='k', lw=2, ls='-', label=expt_label)
        else:
            ax.plot(expt_data[:, 0], expt_data[:, 1], c='k', lw=2, ls='-', label='Experiment')

        if len(profiles) == 1:
            dft_label = 'DFT (this work)'

    for ind, profile in enumerate(profiles):
        if dft_label is None and curve_labels is None:
            stoich_label = get_formula_from_stoich(profile.starting_stoichiometry, tex=True)
        else:
            stoich_label = None

        _label = stoich_label if dft_label is None else dft_label
        if curve_labels is not None and len(curve_labels) > ind:
            _label = curve_labels[ind]

        _line_kwargs = {'c': list(plt.rcParams['axes.prop_cycle'].by_key()['color'])[ind+2]}
        if line_kwargs is not None:
            _line_kwargs.update(line_kwargs[ind])

        _add_voltage_curve(profile.capacities, profile.voltages, ax, label=_label, **_line_kwargs)

    if labels:
        if len(profiles) > 1:
            print("Only labelling first voltage profile.")
        for ind, reaction in enumerate(profiles[0].reactions):
            _labels = []
            for phase in reaction:
                if phase[0] is None or phase[0] == 1.0:
                    _label = ""
                else:
                    _label = "{:.1f} ".format(phase[0])
                _label += "{}".format(phase[1])
                _labels.append(_label)
            _label = '+'.join(_labels)
            _position = (profiles[0].capacities[ind], profiles[0].voltages[ind] + max(profiles[0].voltages)*0.01)
            ax.annotate(_label, xy=_position, textcoords="data", ha="center", zorder=9999)

    if expt or len(profiles) > 1:
        ax.legend()

    ax.set_ylabel('Voltage (V) vs {ion}$^+/${ion}'.format(ion=profile.active_ion))
    ax.set_xlabel('Gravimetric cap. (mAh/g)')

    _, end = ax.get_ylim()
    from matplotlib.ticker import MultipleLocator
    ax.yaxis.set_major_locator(MultipleLocator(0.2))
    ax.set_ylim(0, 1.1 * end)
    _, end = ax.get_xlim()
    ax.set_xlim(0, 1.1 * end)
    ax.grid(False)
    plt.tight_layout(pad=0.0, h_pad=1.0, w_pad=0.2)

    if savefig:
        plt.savefig(savefig)
        print('Wrote {}'.format(savefig))

    elif show:
        plt.show()

    return ax


def _add_voltage_curve(capacities, voltages, ax, label=None, **kwargs):
    """ Add the voltage curves stored under hull['voltage_data'] to the plot.

    Parameters:
        capacities (list): list or numpy array of capacities.
        voltages (list): list or numpy array of voltages.
        ax (matplotlib.axes.Axes): an existing axis object on which to plot.

    Keyword arguments:
        **kwargs (dict): to pass to matplotlib, using abbreviated names (e.g. 'c' not 'color')
        label (str): if present, add this label to the first segment of the voltage curve
            so it can be added to the legend.
        alpha (float): transparency of line

    """
    line_kwargs = {
        'lw': 2,
        'alpha': 1,
    }
    line_kwargs.update(kwargs)

    for i in range(1, len(voltages) - 1):
        if i == 1 and label is not None:
            ax.plot([capacities[i - 1], capacities[i]], [voltages[i], voltages[i]], label=label, **line_kwargs)
        else:
            ax.plot([capacities[i - 1], capacities[i]], [voltages[i], voltages[i]], **line_kwargs)
        if i != len(voltages) - 2:
            ax.plot([capacities[i], capacities[i]], [voltages[i], voltages[i + 1]], **line_kwargs)


@plotting_function
def plot_volume_curve(hull, ax=None, show=True, legend=False, as_percentages=False, **kwargs):
    """ Plot volume curve calculated for phase diagram.

    Parameters:
        hull (matador.hull.QueryConvexHull): matador hull object.

    Keyword arguments:
        ax (matplotlib.axes.Axes): an existing axis on which to plot.
        show (bool): whether or not to display plot in X-window.
        legend (bool): whether to add the legend.
        as_percentages (bool): whether to show the expansion as a percentage.

    """
    import matplotlib.pyplot as plt

    if ax is None:
        if hull.savefig or any([kwargs.get(ext) for ext in SAVE_EXTS]):
            fig = plt.figure()
        else:
            fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        ax = ax

    if as_percentages:
        volume_key = 'volume_expansion_percentage'
    else:
        volume_key = 'volume_ratio_with_bulk'

    for j in range(len(hull.volume_data['electrode_volume'])):
        c = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])[j+2]
        stable_hull_dist = hull.volume_data['hull_distances'][j]
        if len(stable_hull_dist) != len(hull.volume_data['Q'][j]):
            raise RuntimeError("This plot does not support --hull_cutoff.")

        ax.plot(
            [q for ind, q in enumerate(hull.volume_data['Q'][j][:-1]) if stable_hull_dist[ind] == 0],
            [v for ind, v in enumerate(hull.volume_data[volume_key][j]) if stable_hull_dist[ind] == 0],
            marker='o', markeredgewidth=1.5, markeredgecolor='k', c=c, zorder=1000, lw=0,
        )

        ax.plot(
            [q for ind, q in enumerate(hull.volume_data['Q'][j][:-1]) if stable_hull_dist[ind] == 0],
            [v for ind, v in enumerate(hull.volume_data[volume_key][j]) if stable_hull_dist[ind] == 0],
            lw=2, c=c,
            label=("{}"
                   .format(get_formula_from_stoich(hull.volume_data['endstoichs'][j], tex=True)))
        )

    ax.set_xlabel("Gravimetric capacity (mAh/g)")
    if as_percentages:
        ax.set_ylabel('Volume expansion (%)')
    else:
        ax.set_ylabel('Volume ratio with starting electrode')

    if legend or len(hull.volume_data['Q']) > 1:
        ax.legend()
    fname = '{}_volume'.format(''.join(hull.elements))

    if hull.savefig or any([kwargs.get(ext) for ext in SAVE_EXTS]):
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                plt.savefig('{}.{}'.format(fname, ext), transparent=True)
                print('Wrote {}.{}'.format(fname, ext))

    if show:
        plt.show()



# coding: utf-8
# Distributed under the terms of the MIT License.

from .atomic_scattering_loaders import load_scattering_file

__all__ = ["GSAS_ATOMIC_SCATTERING_COEFFS"]


def load_atomic_scattering_factors_gsas():
    """ Load atomic scattering factors used by GSAS-II, from the paper:

    "New Analytical Scattering Factor Functions for Free Atoms and Ions
    for Free Atoms and Ions", D. Waasmaier & A. Kirfel, Acta Cryst.
    (1995). A51, 416-413.

    """
    from pathlib import Path

    filename = Path(__file__).parent.joinpath("atomic_scattering_factors_gsas.dat")
    return load_scattering_file(
        filename, a_inds=(1, 2, 3, 4), b_inds=(6, 7, 8, 9), c_ind=5
    )


GSAS_ATOMIC_SCATTERING_COEFFS = load_atomic_scattering_factors_gsas()



""" This submodule includes functionality to create a
template .matadorrc file.

"""

__all__ = ['quickstart_settings']

import os
from matador.utils.print_utils import print_warning, print_success


def quickstart_settings():
    """ Call this if no user-specified configuration file is found. A series
    of questions will be asked to create a basic matadorrc.

    """
    print('No user-specified input found at either ~/.matadorrc or '
          '~/.config/matadorrc')
    response = input('Would you like to use quickstart to create one now? [y/n] ')
    if not response.lower().startswith('y'):
        return None

    hline = 80*'-'
    print('Okay, let\'s get started!')
    print('Firstly, where would you like me to put your config file: ')
    valid = False
    while not valid:
        response = input('[1] ~/.matadorrc, [2] ~/.config/matadorrc? ')
        if response not in ['1', '2']:
            print('I didn\'t understand, sorry... please enter 1 or 2')
        else:
            valid = True

    if response == '1':
        fname = os.path.expanduser('~/.matadorrc')
    else:
        fname = os.path.expanduser('~/.config/matadorrc')

    flines = []
    print(hline)

    response = input('Great! Would you like to use matador with a MongoDB database? [y/n] ')
    if response.lower().startswith('y'):
        flines.append('mongo:')
        host_response = input('Where is this MongoDB hosted? [default: localhost] ')
        if host_response == '':
            host_response = 'localhost'
        flines.append('  host: {}'.format(host_response))
        print(hline)
        port_response = input('and at what port? [default: 27017] ')
        if port_response == '':
            port_response = 27017
        flines.append('  port: {}'.format(port_response))
        print(hline)
        print('Please set your firewall and MongoDB settings accordingly: '
              'by default MongoDB\nneeds no authentication so be careful '
              'if your database is running on an open port!')
        test_conn_response = input('Would you like me to test the connection to this MongoDB instance now? [y/n] ')
        keep_trying = True
        if test_conn_response.lower() == 'y':
            while keep_trying:
                import pymongo as pm
                cli = pm.MongoClient(host_response,
                                     port=int(port_response),
                                     maxIdleTimeMS=600000,
                                     socketTimeoutMS=3600000,
                                     serverSelectionTimeoutMS=10000,
                                     connectTimeoutMS=10000)
                try:
                    cli.database_names()
                    keep_trying = False
                except pm.errors.ServerSelectionTimeoutError:
                    print_warning('Failed to connect to {}:{}, are you sure it is running?'
                                  .format(host_response, port_response))
                    response = input('Would you like me to try again? [y/n] ')
                    keep_trying = response.lower().startswith('y')
                except Exception:
                    keep_trying = False

        print(hline)
        db_response = input('What is the name of the database (not collection) '
                            'that you want to query/create? [default: crystals] ')
        if db_response == '':
            db_response = 'crystals'
        flines.append('  db: {}'.format(db_response))

        print(hline)
        coll_response = input('What about the default collection name to query? [default: repo] ')
        if coll_response == '':
            coll_response = 'repo'
        flines.append('  default_collection: {}'.format(coll_response))

        print(hline)
        print('Would you like to set a protected folder for this collection?')
        file_path_response = input('Any import to this collection must happen from this path, and the '
                                   '--force flag must be used if so. [default: None] ')
        if file_path_response != '':
            flines.append('  default_file_collection_path: {}'.format(file_path_response))

        print(hline)
        print_success('Mongo section of matadorrc complete! Saving as {} and exiting...'.format(fname))
        print(hline)

    if not flines:
        print('Okay, there is no config file to be made and I will keep using the defaults.')
        return None

    with open(fname, 'a') as f:
        for line in flines:
            f.write(line + '\n')

    return fname



# coding: utf-8
# Distributed under the terms of the MIT License.

from .atomic_scattering_factors_gsas import GSAS_ATOMIC_SCATTERING_COEFFS
from .atomic_scattering_factors_raspa import RASPA_ATOMIC_SCATTERING_COEFFS

__all__ = [
    "RASPA_ATOMIC_SCATTERING_COEFFS",
    "GSAS_ATOMIC_SCATTERING_COEFFS",
]



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements various workflows, ways
of chaining up different calculations at high-throughput.

"""

import abc
import os
import logging
from matador.utils.print_utils import dumps

LOG = logging.getLogger('run3')


class Workflow:
    """ Workflow objects are bundles of calculations defined as
    :obj:`WorkflowStep` objects. Each :obj:`WorkflowStep` takes three arguments:
    the :obj:`matador.compute.ComputeTask` object used to run the calculations, the calculation
    parameters (which can be modified by each step), the seed name.
    Any subclass of Workflow must implement `preprocess` and `postprocess`
    methods (even if they just return True).

    Attributes:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be running the computation.
        calc_doc (dict): the interim dictionary of structural and
            calculation parameters.
        seed (str): the root seed for the calculation.
        label (str): the name of the type of the Workflow object.
        success (bool): the status of the workflow. This is only set to True after
            final step completes, but BEFORE post-processing.
        steps (:obj:`list` of :obj:`WorkflowStep`): list of steps to be
            completed.

    """
    def __init__(self, relaxer, calc_doc, seed, **workflow_kwargs):
        """ Initialise the Workflow object from a :obj:`matador.compute.ComputeTask`, calculation
        parameters and the seed name.

        Parameters:
            relaxer (:obj:`matador.compute.ComputeTask`): the object that will be running the computation.
            calc_doc (dict): dictionary of structure and calculation
                parameters.
            seed (str): root seed for the calculation.

        Raises:
            RuntimeError: if any part of the calculation fails.

        """
        self.relaxer = relaxer
        self.calc_doc = calc_doc
        self.seed = seed
        self.label = self.__class__.__name__
        self.compute_dir = relaxer.compute_dir
        self.success = None
        self.steps = []
        self.clean_after_step = []
        self.workflow_params = workflow_kwargs

        LOG.info('Performing Workflow of type {} on {}'.format(self.label, self.seed))

        self.preprocess()
        try:
            self.run_steps()
        except RuntimeError as exc:
            LOG.critical('Workflow failed: calling postprocess()')
            self._clean_up()
            raise exc

        if self.success:
            self.postprocess()

        self._clean_up()

    @abc.abstractmethod
    def preprocess(self):
        """ This function is run at the start of the workflow, and is
        responsible for adding WorkflowStep objects to the Workflow.

        """
        raise NotImplementedError('Please implement a preprocess method.')

    def postprocess(self):
        """ This OPTIONAL function is run upon successful completion of all steps
        of the workflow and can be overloaded by the subclass to perform
        any postprocessing steps. This occurs *before* cleaning up the
        directory (i.e. moving to completed/bad_castep).

        """

    def _clean_up(self, success=None):
        """ This method moves files to `completed/` or `bad_castep/` depending on the status
        of the workflow. It will use the current seed of the relaxer, so this function can be
        called at intermediate steps if this seed changes.

        """
        cwd = os.getcwd()
        if success is None:
            success = self.success

        if self.compute_dir:
            os.chdir(self.compute_dir)
            if success:
                LOG.info('Writing results from compute dir of Workflow {} run to completed folder and tidying up.'
                         .format(self.label))
                self.relaxer.mv_to_completed(self.relaxer.seed, keep=True, skip_existing=False)
            else:
                LOG.info('Writing results from compute dir of failed Workflow {} run to bad_castep folder and tidying up.'
                         .format(self.label))
                self.relaxer.mv_to_bad(self.relaxer.seed)
            os.chdir(cwd)

        if success:
            LOG.info('Writing results of Workflow {} run to completed folder and tidying up.'.format(self.label))
            self.relaxer.mv_to_completed(self.relaxer.seed, keep=True, skip_existing=True)
        else:
            LOG.info('Writing results of failed Workflow {} run to bad_castep folder and tidying up.'.format(self.label))
            self.relaxer.mv_to_bad(self.relaxer.seed)

    def add_step(self, function, name, input_exts=None, output_exts=None, clean_after=False, **func_kwargs):
        """ Add a step to the workflow.

        Parameters:
            function (Function): the function to run in the step; must
                accept arguments of (self.relaxer, self.calc_doc, self.seed).
            name (str): the desired name for the step (human-readable).

        Keyword arguments:
            clean_after (bool): whether or not to clean up after this step is called
            func_kwargs (dict): any arguments to pass to function when called.

        """
        self.steps.append(WorkflowStep(function, name,
                                       self.compute_dir, input_exts, output_exts,
                                       **func_kwargs))
        self.clean_after_step.append(clean_after)

    def run_steps(self):
        """ Loop over steps and run them. """
        try:
            if not self.steps:
                msg = 'No steps added to Workflow!'
                LOG.error(msg)
                raise RuntimeError(msg)

            for ind, step in enumerate(self.steps):
                LOG.info("Running step {step.name}: {step.function}".format(step=step))
                LOG.debug("Current state: " + dumps(self.calc_doc, indent=None))
                success = step.run_step(self.relaxer, self.calc_doc, self.seed)
                if self.clean_after_step[ind]:
                    self._clean_up(success=success)

            self.success = True

        except RuntimeError:
            self.success = False
            msg = '{} workflow exiting...'.format(self.label)
            LOG.error(msg)
            raise RuntimeError(msg)


class WorkflowStep:
    """ An individual step in a Workflow, defined by a Python function
    and a name. The function will be called with arguments
    (relaxer, calc_doc, seed) with the run_step method.

    Attributes:
        function (function): the function to call.
        name (str): the human-readable name of the step.
        compute_dir (str): the folder that relaxer will perform the calculation in.
        func_kwargs (dict): any extra kwargs to pass to the function.
        input_exts (list): list of input file extensions to cache after running.
        output_exts (list): list of output file extensions to cache after running.

    """

    success = False

    def __init__(self, function, name, compute_dir=None, input_exts=None, output_exts=None, **func_kwargs):
        """ Construct a WorkflowStep from a function. """
        LOG.debug('Constructing WorkflowStep: {}'.format(name))
        self.function = function
        self.name = name
        self.compute_dir = compute_dir
        self.func_kwargs = func_kwargs
        self.input_exts = input_exts
        self.output_exts = output_exts

    def _cache_files(self, seed, exts, mode, directory=None):
        """ Copy any files <seed>.<ext> for ext in exts to
        <seed>.<ext>_<label>.

        Parameters:
            seed (str): seed for the workflow step.
            exts (:obj:`list` of :obj:`str`): list of file extensions, including '.'.
            mode (str): either 'in' (warning printed if file missing) or 'out' (no warning).

        """
        import shutil
        import glob

        for ext in exts:
            if '*' in ext:
                srcs = glob.glob('{}{}'.format(seed, ext))
            else:
                srcs = ['{}{}'.format(seed, ext)]
            for src in srcs:
                dst = src + '_{}'.format(self.name)
                if os.path.isfile(src):
                    shutil.copy2(src, dst, follow_symlinks=True)
                    LOG.info('Backed up {} file {} to {}.'.format(mode, src, dst))
                else:
                    if mode == 'in':
                        error = 'Failed to cache input file {} for step {}.'.format(src, self.name)
                        LOG.warning(error)

    def _cache_inputs(self, seed):
        """ Save any input files for the WorkflowStep with appropriate suffix
        as determined by the WorkflowStep label. All files with <seed>.<ext>
        will be moved to <seed>.<ext>_<name>, for any <ext> inside the
        `input_exts` attribute. This is called after the WorkflowStep has
        finished, even if it does not succeed...

        Parameters:
            seed (str): seed for the workflow step.

        """
        if self.input_exts is not None:
            self._cache_files(seed, self.input_exts, 'in')

    def _cache_outputs(self, seed):
        """ Save any output files for the WorkflowStep with appropriate suffix
        as determined by the WorkflowStep label. All files with <seed>.<ext>
        will be moved to <seed>.<ext>_<name>, for any <ext> inside the
        `output_exts` attribute.

        Parameters:
            seed (str): seed for the workflow step.

        """
        if self.output_exts is not None:
            self._cache_files(seed, self.output_exts, 'out')

    def cache_files(self, seed):
        """ Wrapper for calling both _cache_inputs and _cache_outputs, without
        throwing any errors.
        """
        cwd = os.getcwd()
        if self.compute_dir is not None:
            os.chdir(self.compute_dir)

        self._cache_inputs(seed)
        self._cache_outputs(seed)

        if self.compute_dir is not None:
            os.chdir(cwd)

    def run_step(self, relaxer, calc_doc, seed):
        """ Run the workflow step.

        Parameters:
            relaxer (:obj:`matador.compute.ComputeTask`): the object that will be running the computation.
            calc_doc (dict): dictionary of structure and calculation
                parameters.
            seed (str): root seed for the calculation.

        Raises:
            RuntimeError: if any step fails.

        """
        try:
            LOG.info('WorkflowStep {} starting...'.format(self.name))
            self.success = self.function(relaxer, calc_doc, seed, **self.func_kwargs)
        except RuntimeError as exc:
            msg = 'WorkflowStep {} failed with error {}.'.format(self.name, exc)
            LOG.error(msg)
            self.success = False
            self.cache_files(seed)
            raise exc

        if self.success is None:
            LOG.info('WorkflowStep {} skipped, did you provide all the input files?'.format(self.name))
            return self.success

        if self.success:
            LOG.info('WorkflowStep {} completed successfully.'.format(self.name))
        else:
            LOG.warning('WorkflowStep {} was unsuccessful.'.format(self.name))

        self.cache_files(seed)

        return self.success



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the CastepElasticWorkflow, which performs
elastic tensor calculations with CASTEP. So far, only bulk modulus
calculations (i.e. varying volume from equilibrium) have been implemented.

"""


import copy
import logging
import numpy as np
from matador.workflows import Workflow
from matador.crystal.elastic import get_equation_of_state

LOG = logging.getLogger('run3')


def castep_elastic(relaxer, calc_doc, seed, **kwargs):
    """ Perform a calculation of the elastic tensor on a system.
    Currently only calculation of the bulk modulus is implemented.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): dictionary of structure and calculation
            parameters.
        seed (str): root seed for the calculation.

    Raises:
        RuntimeError: if any part of the calculation fails.

    Returns:
        bool: True if Workflow completed successfully, or False otherwise.

    """
    workflow = CastepElasticWorkflow(relaxer, calc_doc, seed, **kwargs)
    return workflow.success


class CastepElasticWorkflow(Workflow):
    """ Perform a "full" spectral calculation on a system, i.e. first
    perform an SCF then interpolate to different kpoint paths/grids to
    form DOS and dispersions. Optionally use OptaDOS for post-processing
    of DOS.

    Attributes:
        relaxer (:obj:`ComputeTask`): the object that calls CASTEP.
        calc_doc (dict): the interim dictionary of structural and
            calculation parameters.
        seed (str): the root seed for the calculation.
        volume_rescale (list): a list of floats containing the amounts
            by which to rescale volumes when performing bulk modulus calculation.
        success (bool): the status of the Workflow: only set to True after
            post-processing method completes.

    """

    def preprocess(self):
        """ Decide which parts of the Workflow need to be performed,
        and set the appropriate CASTEP parameters.

        """
        num_volumes = self.workflow_params.get('num_volumes', 9)
        self.plot = self.workflow_params.get('plot', True)

        self.volume_rescale = np.cbrt(np.geomspace(0.7, 1.2, num=num_volumes, endpoint=True)).tolist()
        self.volume_rescale.append(1.0)
        self._completed_volumes = []
        LOG.info('Preprocessing completed: run3 bulk modulus calculation options {}'
                 .format(self.volume_rescale))

        # clean up after geometry step as seed is going to change
        self.add_step(castep_elastic_prerelax, 'relax', clean_after=True)
        for volume in self.volume_rescale:
            self.add_step(castep_rescaled_volume_scf, 'rescaled_volume_scf', rescale=volume)

    def postprocess(self):
        """ Fit some equations of state, then save a plot and a datafile. """
        results = get_equation_of_state(self.seed + '.bulk_mod', plot=self.plot)
        if 'summary' in results:
            for line in results['summary']:
                print(line)
            print('Writing summary to {seed}.bulk_mod.results'.format(seed=self.seed))
            with open(self.seed + '.bulk_mod.results', 'w') as f:
                f.writelines(results['summary'])


def castep_rescaled_volume_scf(relaxer, calc_doc, seed, rescale=1):
    """ Run a singleshot SCF calculation.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    Keyword arguments:
        rescale (float): amount to rescale volume by.

    """
    assert rescale > 0
    LOG.info('Performing CASTEP SCF on volume rescaled by {:.2f}.'.format(rescale**3))
    scf_doc = copy.deepcopy(calc_doc)
    for i in range(3):
        for k in range(3):
            scf_doc['lattice_cart'][i][k] *= rescale
    scf_doc['task'] = 'singlepoint'
    bulk_mod_seed = seed + '.bulk_mod'
    relaxer.seed = bulk_mod_seed

    return relaxer.run_castep_singleshot(scf_doc, bulk_mod_seed, keep=True, intermediate=True)


def castep_elastic_prerelax(relaxer, calc_doc, seed):
    """ Run a geometry optimisation before re-scaling volumes SCF-style calculation.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP elastic pre-relax...')
    relax_doc = copy.deepcopy(calc_doc)
    if 'geom_max_iter' not in relax_doc:
        relax_doc['geom_max_iter'] = 100
    relax_doc['task'] = 'geometryoptimisation'

    required = []
    forbidden = []
    relaxer.validate_calc_doc(relax_doc, required, forbidden)
    relaxer.calc_doc = relax_doc

    return relaxer.run_castep_relaxation(intermediate=True)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the CastepPhononWorkflow,
which performs phonon calculations with CASTEP in
multiple steps (only when necessary):

    1. Try to pre-relax structure (skipped if check file
           is already present).
    2. Calculate dynamical matrix.
    3. If `phonon_fine_kpoint_mp_spacing` keyword is found,
       interpolate dynamical matrix to form phonon DOS.
    4. If `phonon_fine_kpoint_path_spacing` key word is found,
       interpolate dynamical matrix to form phonon dispersion
       on the path given by seekpath.
    5. If `task=thermodynamics`, perform a CASTEP thermodynamics
       calculation for the temperature-dependence of the free energy.

"""


import os
import copy
import logging
from matador.workflows.workflows import Workflow

LOG = logging.getLogger('run3')


def castep_full_phonon(relaxer, calc_doc, seed, **kwargs):
    """ Perform a "full" phonon calculation on a system, i.e.
    first perform a relaxation in a standardised unit cell,
    then compute the dynamical matrix, then finally interpolate
    that dynamical matrix into dispersion curves and DOS. This function
    is a wrapper for the CastepPhononWorkflow class.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): dictionary of structure and calculation
            parameters.
        seed (str): root seed for the calculation.

    Raises:
        RuntimeError: if any part of the calculation fails.

    Returns:
        bool: True if Workflow completed successfully, or False otherwise.

    """
    workflow = CastepPhononWorkflow(relaxer, calc_doc, seed, **kwargs)
    return workflow.success


class CastepPhononWorkflow(Workflow):
    """ Perform a "full" phonon calculation on a system, i.e.
    first perform a relaxation in a standardised unit cell,
    then compute the dynamical matrix, then finally interpolate
    that dynamical matrix into dispersion curves and DOS.

    Attributes:
        relaxer (:obj:`ComputeTask`): the object that calls CASTEP.
        calc_doc (dict): the interim dictionary of structural and
            calculation parameters.
        seed (str): the root seed for the calculation.
        success (bool): the status of the Workflow: only set to True after
            post-processing method completes.

    """
    def preprocess(self):
        """ Decide which parts of the Workflow need to be performed,
        and set the appropriate CASTEP parameters.

        """
        # default todo
        todo = {'relax': True, 'dynmat': True, 'vdos': False, 'dispersion': False, 'thermodynamics': False}
        # definition of steps and names
        steps = {'relax': castep_phonon_prerelax,
                 'dynmat': castep_phonon_dynmat,
                 'vdos': castep_phonon_dos,
                 'dispersion': castep_phonon_dispersion,
                 'thermodynamics': castep_phonon_thermodynamics}

        exts = {
            'relax':
                {'input': ['.cell', '.param'], 'output': ['.castep', '-out.cell', '.*err']},
            'dynmat':
                {'input': ['.cell', '.param'], 'output': ['.castep', '.*err']},
            'vdos':
                {'input': ['.cell', '.param'], 'output': ['.castep', '.phonon', '.phonon_dos', '.*err']},
            'dispersion':
                {'input': ['.cell', '.param'], 'output': ['.castep', '.phonon', '.*err']},
            'thermodynamics':
                {'input': ['.cell', '.param'], 'output': ['.castep', '.*err']}
        }

        if self.calc_doc.get('task').lower() in ['phonon', 'thermodynamics']:
            if ('phonon_fine_kpoint_path' in self.calc_doc or 'phonon_fine_kpoint_list' in self.calc_doc
                    or 'phonon_fine_kpoint_path_spacing' in self.calc_doc):
                todo['dispersion'] = True
            if 'phonon_fine_kpoint_mp_spacing' in self.calc_doc:
                todo['vdos'] = True
            if self.calc_doc['task'].lower() == 'thermodynamics':
                todo['thermodynamics'] = True

        # prepare to do pre-relax if there's no check file
        if os.path.isfile(self.seed + '.check'):
            todo['relax'] = False
            LOG.info('Restarting from {}.check, so not performing re-relaxation'.format(self.seed))

        for key in todo:
            if todo[key]:
                self.add_step(steps[key], key,
                              input_exts=exts[key].get('input'),
                              output_exts=exts[key].get('output'))

        # always standardise the cell so that any phonon calculation can have
        # post-processing performed after the fact, unless a path has been provided
        if 'phonon_fine_kpoint_list' not in self.calc_doc and 'phonon_fine_kpoint_path' not in self.calc_doc:
            from matador.utils.cell_utils import cart2abc
            prim_doc, kpt_path = self.relaxer.get_seekpath_compliant_input(
                self.calc_doc, self.calc_doc.get('phonon_fine_kpoint_path_spacing', 0.02))
            self.calc_doc.update(prim_doc)
            self.calc_doc['lattice_abc'] = cart2abc(self.calc_doc['lattice_cart'])
            if todo['dispersion']:
                self.calc_doc['phonon_fine_kpoint_list'] = kpt_path

        elif todo['dispersion'] and 'phonon_fine_kpoint_path' in self.calc_doc:
            self._user_defined_kpt_path = True
            LOG.warning('Using user-defined k-point path for all structures.')
            self.calc_doc['phonon_fine_kpoint_spacing'] = self.calc_doc.get('phonon_fine_kpoint_path_spacing', 0.05)

        # always shift phonon grid to include Gamma
        if 'phonon_kpoint_mp_spacing' in self.calc_doc:
            from matador.utils.cell_utils import calc_mp_grid, shift_to_include_gamma
            grid = calc_mp_grid(self.calc_doc['lattice_cart'], self.calc_doc['phonon_kpoint_mp_spacing'])
            offset = shift_to_include_gamma(grid)
            if offset != [0, 0, 0]:
                self.calc_doc['phonon_kpoint_mp_offset'] = offset
                LOG.debug('Set phonon MP grid offset to {}'.format(offset))

        LOG.info('Preprocessing completed: run3 phonon options {}'.format(todo))


def castep_phonon_prerelax(relaxer, calc_doc, seed):
    """ Run a singleshot geometry optimisation before an SCF-style calculation.
    This is typically used to ensure phonon calculations start successfully.
    The phonon calculation will then be restarted from the .check file produced here.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP phonon pre-relax...')
    relax_doc = copy.deepcopy(calc_doc)
    relax_doc['write_checkpoint'] = 'ALL'
    if 'geom_max_iter' not in relax_doc:
        relax_doc['geom_max_iter'] = 100
    relax_doc['task'] = 'geometryoptimisation'

    required = []
    forbidden = ['phonon_fine_kpoint_list',
                 'phonon_fine_kpoint_path',
                 'phonon_fine_kpoint_mp_spacing',
                 'phonon_fine_kpoint_path_spacing']

    relaxer.validate_calc_doc(relax_doc, required, forbidden)
    relaxer.calc_doc = relax_doc

    relaxer.run_castep_relaxation(intermediate=True)


def castep_phonon_dynmat(relaxer, calc_doc, seed):
    """ Runs a singleshot phonon dynmat calculation, with no "fine_method" interpolation.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP dynmat calculation...')
    dynmat_doc = copy.deepcopy(calc_doc)
    dynmat_doc['write_checkpoint'] = 'ALL'
    dynmat_doc['task'] = 'phonon'

    required = []
    forbidden = ['phonon_fine_kpoint_list',
                 'phonon_fine_kpoint_path',
                 'phonon_fine_kpoint_mp_spacing',
                 'phonon_fine_kpoint_path_spacing']

    relaxer.validate_calc_doc(dynmat_doc, required, forbidden)
    return relaxer.run_castep_singleshot(dynmat_doc, seed, keep=True, intermediate=True)


def castep_phonon_dos(relaxer, calc_doc, seed):
    """ Runs a DOS interpolation on top of a completed
    phonon calculation.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP phonon DOS calculation...')
    dos_doc = copy.deepcopy(calc_doc)
    dos_doc['task'] = 'phonon'
    dos_doc['phonon_calculate_dos'] = True
    dos_doc['continuation'] = 'default'

    required = ['phonon_fine_kpoint_mp_spacing']
    forbidden = ['phonon_fine_kpoint_list',
                 'phonon_fine_kpoint_path',
                 'phonon_fine_kpoint_path_spacing']

    relaxer.validate_calc_doc(dos_doc, required, forbidden)

    return relaxer.run_castep_singleshot(dos_doc, seed, keep=True, intermediate=True)


def castep_phonon_dispersion(relaxer, calc_doc, seed):
    """ Runs a dispersion interpolation on top of a completed
    phonon calculation.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP phonon dispersion calculation...')
    disp_doc = copy.deepcopy(calc_doc)
    disp_doc['task'] = 'phonon'
    disp_doc['phonon_calculate_dos'] = False
    disp_doc['continuation'] = 'default'

    required = []
    forbidden = ['phonon_fine_kpoint_mp_spacing']

    relaxer.validate_calc_doc(disp_doc, required, forbidden)

    return relaxer.run_castep_singleshot(disp_doc, seed, keep=True, intermediate=True)


def castep_phonon_thermodynamics(relaxer, calc_doc, seed):
    """ Runs a "thermodynamics" interpolation on top of a completed
    phonon calculation, using the phonon_fine_kpoint_mp_grid.

    Parameters:
        relaxer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP thermodynamics calculation...')
    thermo_doc = copy.deepcopy(calc_doc)
    thermo_doc['continuation'] = 'default'
    thermo_doc['task'] = 'thermodynamics'
    thermo_doc['phonon_calculate_dos'] = False

    required = ['phonon_fine_kpoint_mp_spacing']
    forbidden = ['phonon_fine_kpoint_list',
                 'phonon_fine_kpoint_path',
                 'phonon_fine_kpoint_path_spacing']

    relaxer.validate_calc_doc(thermo_doc, required, forbidden)

    return relaxer.run_castep_singleshot(thermo_doc, seed, keep=True, intermediate=True)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" Some simple utilities for making DB connections. """


import pymongo as pm
from matador.config import load_custom_settings


def make_connection_to_collection(coll_names, check_collection=False, allow_changelog=False,
                                  mongo_settings=None, override=False, import_mode=False, quiet=True, debug=False):
    """ Connect to database of choice.

    Parameters:
        coll_names (str): name of collection.

    Keyword Arguments:
        check_collection (bool): check whether collections exist (forces connection)
        allow_changelog (bool): allow queries to collections with names prefixed by __
        mongo_settings (dict): dict containing mongo and related config
        override (bool): don't ask for user input from stdin and assume all is well
        quiet (bool): don't print very much.

    Returns:
        client (MongoClient): the connection to the database
        db (Database): the database to query
        collections (dict): Collection objects indexed by name

    """

    if mongo_settings is None:
        settings = load_custom_settings(no_quickstart=override)
    else:
        settings = mongo_settings

    if not quiet:
        print('Trying to connect to {host}:{port}/{db}'.format(**settings['mongo']))

    client = pm.MongoClient(
        host=settings['mongo']['host'],
        port=settings['mongo']['port'],
        connect=False,
        maxIdleTimeMS=600000,  # disconnect after 10 minutes idle
        socketTimeoutMS=3600000,  # give up on database after 1 hr without results
        serverSelectionTimeoutMS=10000,  # give up on server after 2 seconds without results
        connectTimeoutMS=10000)  # give up trying to connect to new database after 2 seconds

    try:
        database_names = client.list_database_names()
        if not quiet:
            print('Success!')
    except pm.errors.ServerSelectionTimeoutError as exc:
        print('{}: {}'.format(type(exc).__name__, exc))
        raise SystemExit('Unable to connect to {host}:{port}/{db}, exiting...'.format(**settings['mongo']))

    if settings['mongo']['db'] not in database_names:
        if override:
            response = 'y'
        else:
            response = input('Database {db} does not exist at {host}:{port}/{db}, '
                             'would you like to create it? (y/n) '
                             .format(**settings['mongo']))

        if response.lower() != 'y':
            raise SystemExit('Exiting...')
        else:
            print('Creating database {}'.format(settings['mongo']['db']))

    db = client[settings['mongo']['db']]
    possible_collections = [name for name in db.list_collection_names() if not name.startswith('__')]
    collections = dict()
    # allow lists of collections for backwards-compat, though normally
    # we only want to connect to one at a time
    if coll_names is not None:
        if not isinstance(coll_names, list):
            coll_names = [coll_names]

        if len(coll_names) > 1:
            raise NotImplementedError("Querying multiple collections is no longer supported.")

        for collection in coll_names:

            if not allow_changelog:
                if collection.startswith('__'):
                    raise SystemExit('Queries to collections prefixed with __ are VERBOTEN!')

                if collection not in possible_collections:
                    options = fuzzy_collname_match(collection, possible_collections)
                    if not options and check_collection:
                        client.close()
                        raise SystemExit('Collection {} not found!'.format(collection))
                    else:
                        print('Collection {} not found, did you mean one of these?'.format(collection))
                        for ind, value in enumerate(options[:10]):
                            print('({}):\t{}'.format(ind, value))
                        if check_collection:
                            try:
                                choice = int(input('Please enter your choice: '))
                                collection = options[choice]
                            except Exception:
                                raise SystemExit('Invalid choice. Exiting...')
                        elif import_mode:
                            if override:
                                choice = 'y'
                            else:
                                choice = input('Are you sure you want to make a new collection called {}? (y/n) '
                                               .format(collection))
                            if choice.lower() != 'y' and choice.lower != 'yes':
                                try:
                                    choice = int(input('Then please enter your choice from above: '))
                                    collection = options[choice]
                                except Exception:
                                    raise SystemExit('Invalid choice. Exiting...')

            collections[collection] = db[collection]
    else:
        default_collection = settings['mongo']['default_collection']
        if default_collection not in possible_collections:
            if check_collection:
                client.close()
                raise SystemExit('Default collection {} not found!'.format(default_collection))
            else:
                print('Creating new collection {}...'.format(default_collection))
        collections['repo'] = db[default_collection]

    return client, db, collections


def fuzzy_collname_match(trial, targets):
    """ Do a noddy fuzzy match for bits between punctuation, e.g.
    matthews_cool_database will search for matthews, cool and database
    in the known collection names.

    Parameters:
        trial (str): database search name.
        targets (list): list of existing database names.

    Returns:
        list: list of roughly matching collection names ordered
            by occurence of tokens.

    """
    split_chars = set([char for char in trial if (not char.isalpha() and not char.isdigit())])
    tokens = trial
    for char in split_chars:
        tokens = tokens.replace(char, ' ')
    tokens = tokens.split()
    options = {}
    for token in tokens:
        for option in targets:
            if token in option:
                if option not in options:
                    options[option] = 0
                else:
                    options[option] += 1
    options_list = [item[0] for item in list(sorted(options.items(), reverse=True, key=lambda item: item[1]))]
    return options_list



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot convex hulls and phase
diagrams generally.

"""


import numpy as np
from matador.utils.chem_utils import get_stoich_from_formula, get_formula_from_stoich
from matador.plotting.plotting import plotting_function, get_linear_cmap, SAVE_EXTS

EPS = 1e-12

__all__ = ['plot_2d_hull', 'plot_ternary_hull', 'plot_ensemble_hull', 'plot_temperature_hull']


def _get_hull_labels(hull, label_cutoff=None, num_species=None, exclude_edges=True):
    """ Return list of structures to labels on phase diagram.

    Parameters:
        hull (matador.hull.QueryConvexHull): phase diagram to plot.

    Keyword arguments:
        num_species (int): structures containing this number of species
            will be labelled.
        exclude_edges (bool): ignore any structure that has zero of any
            chemical potential (i.e the edge of the hull).

    Returns:
        label_cursor (list(dict)): list of matador documents to label.

    """
    eps = 1e-9
    if label_cutoff is None:
        label_cutoff = 0.0
    if isinstance(label_cutoff, list) and len(label_cutoff) == 2:
        label_cutoff = sorted(label_cutoff)
        # first, only apply upper limit as we need to filter by stoich aftewards
        label_cursor = [doc for doc in hull.cursor if doc['hull_distance'] <= label_cutoff[1]]
    else:
        if isinstance(label_cutoff, list):
            assert len(label_cutoff) == 1, 'Incorrect number of label_cutoff values passed, should be 1 or 2.'
            label_cutoff = label_cutoff[0]
        label_cursor = [doc for doc in hull.cursor if doc['hull_distance'] <= label_cutoff + eps]

    num_labels = len({get_formula_from_stoich(doc['stoichiometry']) for doc in label_cursor})
    if num_labels < len(label_cursor):
        tmp_cursor = []
        for doc in label_cursor:
            if doc['stoichiometry'] not in [_doc['stoichiometry'] for _doc in tmp_cursor]:
                tmp_cursor.append(doc)
            else:
                label_cursor = tmp_cursor
    if isinstance(label_cutoff, list) and len(label_cutoff) == 2:
        # now apply lower limit
        label_cursor = [doc for doc in label_cursor if label_cutoff[0] <= doc['hull_distance'] <= label_cutoff[1]]
    # remove chemical potentials and unwanted e.g. binaries
    if num_species is not None:
        label_cursor = [doc for doc in label_cursor if len(doc['stoichiometry']) == num_species]
    if exclude_edges:
        label_cursor = [doc for doc in label_cursor if (all(doc['concentration']) > 0 and
                                                        sum(doc['concentration']) <= 1 - 1e-6)]

    label_cursor = sorted(label_cursor, key=lambda doc: doc['concentration'])

    return label_cursor


@plotting_function
def plot_2d_hull(hull, ax=None, show=True, plot_points=True, plot_tie_line=True,
                 plot_hull_points=True, labels=None, label_cutoff=None, colour_by_source=False,
                 sources=None, hull_label=None, source_labels=None, title=True, plot_fname=None, show_cbar=True,
                 label_offset=(1.15, 0.05), eform_limits=None, legend_kwargs=None,
                 **kwargs):
    """ Plot calculated hull, returning ax and fig objects for further editing.

    Parameters:
        hull (matador.hull.QueryConvexHull): matador hull object.

    Keyword arguments:
        ax (matplotlib.axes.Axes): an existing axis on which to plot,
        show (bool): whether or not to display the plot in an X window,
        plot_points (bool): whether or not to display off-hull structures,
        plot_hull_points (bool): whether or not to display on-hull structures,
        labels (bool): whether to label formulae of hull structures, also read from
            hull.args.
        label_cutoff (float/:obj:`tuple` of :obj:`float`): draw labels less than or
            between these distances form the hull, also read from hull.args.
        colour_by_source (bool): plot and label points by their sources
        alpha (float): alpha value of points when colour_by_source is True
        sources (list): list of possible provenances to colour when colour_by_source
            is True (others will be grey)
        title (str/bool): whether to include a plot title.
        png/pdf/svg (bool): whether or not to write the plot to a file.
        plot_fname (str): filename to write plot to, without file extension.

    Returns:
        matplotlib.axes.Axes: matplotlib axis with plot.

    """
    import matplotlib.pyplot as plt
    import matplotlib.colors as colours

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    if not hasattr(hull, 'colours'):
        hull.colours = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])
    hull.default_cmap_list = get_linear_cmap(hull.colours[1:4], list_only=True)
    hull.default_cmap = get_linear_cmap(hull.colours[1:4], list_only=False)

    if labels is None:
        labels = hull.args.get('labels', False)
    if label_cutoff is None:
        label_cutoff = hull.args.get('label_cutoff')

    scale = 1
    scatter = []
    chempot_labels = [get_formula_from_stoich(get_stoich_from_formula(species, sort=False), tex=True)
                      for species in hull.species]
    tie_line = hull.convex_hull.points[hull.convex_hull.vertices]

    # plot hull structures
    if plot_hull_points:
        ax.scatter(tie_line[:, 0], tie_line[:, 1],
                   c=hull.colours[1],
                   marker='o', zorder=99999, edgecolor='k',
                   s=scale*40, lw=1.5)
        if plot_tie_line:
            ax.plot(np.sort(tie_line[:, 0]), tie_line[np.argsort(tie_line[:, 0]), 1],
                    c=hull.colours[0], zorder=1, label=hull_label,
                    marker='o', markerfacecolor=hull.colours[0],
                    markeredgecolor='k', markeredgewidth=1.5, markersize=np.sqrt(scale*40))
    if plot_tie_line:
        ax.plot(np.sort(tie_line[:, 0]), tie_line[np.argsort(tie_line[:, 0]), 1],
                c=hull.colours[0], zorder=1, label=hull_label, markersize=0)

    if hull.hull_cutoff > 0:
        ax.plot(np.sort(tie_line[:, 0]), tie_line[np.argsort(tie_line[:, 0]), 1] + hull.hull_cutoff,
                '--', c=hull.colours[1], alpha=0.5, zorder=1, label='')

    # annotate hull structures
    if labels or label_cutoff is not None:
        label_cursor = _get_hull_labels(hull, num_species=2, label_cutoff=label_cutoff)
        already_labelled = []
        for ind, doc in enumerate(label_cursor):
            formula = get_formula_from_stoich(doc['stoichiometry'], sort=True)
            if formula not in already_labelled:
                arrowprops = dict(arrowstyle="-|>", lw=2, alpha=1, zorder=1, shrinkA=2, shrinkB=4)
                min_comp = tie_line[np.argmin(tie_line[:, 1]), 0]
                e_f = label_cursor[ind]['formation_' + str(hull.energy_key)]
                conc = label_cursor[ind]['concentration'][0]
                if conc < min_comp:
                    position = (0.8 * conc, label_offset[0] * (e_f - label_offset[1]))
                elif label_cursor[ind]['concentration'][0] == min_comp:
                    position = (conc, label_offset[0] * (e_f - label_offset[1]))
                else:
                    position = (min(1.1 * conc + 0.15, 0.95), label_offset[0] * (e_f - label_offset[1]))
                ax.annotate(get_formula_from_stoich(doc['stoichiometry'],
                                                    latex_sub_style=r'\mathregular',
                                                    tex=True,
                                                    elements=hull.species,
                                                    sort=False),
                            xy=(conc, e_f),
                            xytext=position,
                            textcoords='data',
                            ha='right',
                            va='bottom',
                            arrowprops=arrowprops,
                            zorder=1)
                already_labelled.append(formula)

    # points for off hull structures; we either colour by source or by energy
    if plot_points and not colour_by_source:

        if hull.hull_cutoff == 0:
            # if no specified hull cutoff, ignore labels and colour by hull distance
            cmap = hull.default_cmap
            if plot_points:
                scatter = ax.scatter(hull.structures[np.argsort(hull.hull_dist), 0][::-1],
                                     hull.structures[np.argsort(hull.hull_dist), -1][::-1],
                                     s=scale*40,
                                     c=np.sort(hull.hull_dist)[::-1],
                                     zorder=10000,
                                     cmap=cmap, norm=colours.LogNorm(0.02, 2))

                if show_cbar:
                    cbar = plt.colorbar(scatter, aspect=30, pad=0.02,
                                        ticks=[0, 0.02, 0.04, 0.08, 0.16, 0.32, 0.64, 1.28])
                    cbar.ax.tick_params(length=0)
                    cbar.ax.set_yticklabels([0, 0.02, 0.04, 0.08, 0.16, 0.32, 0.64, 1.28])
                    cbar.ax.yaxis.set_ticks_position('right')
                    cbar.ax.set_frame_on(False)
                    cbar.outline.set_visible(False)
                    cbar.set_label('Distance from hull (eV/atom)')

        elif hull.hull_cutoff != 0:
            # if specified hull cutoff colour those below
            c = hull.colours[1]
            for ind in range(len(hull.structures)):
                if hull.hull_dist[ind] <= hull.hull_cutoff or hull.hull_cutoff == 0:
                    if plot_points:
                        scatter.append(ax.scatter(hull.structures[ind, 0], hull.structures[ind, 1],
                                                  s=scale*40,
                                                  alpha=0.9, c=c,
                                                  zorder=300))
            if plot_points:
                ax.scatter(hull.structures[1:-1, 0], hull.structures[1:-1, 1],
                           s=scale*30, lw=0,
                           alpha=0.3, c=hull.colours[-2],
                           edgecolor='k', zorder=10)

    elif colour_by_source:
        _scatter_plot_by_source(
            hull, ax, scale, kwargs,
            sources=sources, source_labels=source_labels,
            plot_hull_points=plot_hull_points, legend_kwargs=legend_kwargs)

    if eform_limits is None:
        eform_limits = (np.min(hull.structures[:, 1]), np.max(hull.structures[:, 1]))
        lims = (-0.1 if eform_limits[0] >= 0 else 1.4*eform_limits[0],
                eform_limits[1] if eform_limits[0] >= 0 else 0.1)
    else:
        lims = sorted(eform_limits)
    ax.set_ylim(lims)

    if isinstance(title, bool) and title:
        if hull._non_elemental:
            ax.set_title(r'({d[0]})$_\mathrm{{x}}$({d[1]})$_\mathrm{{1-x}}$'.format(d=chempot_labels))
        else:
            ax.set_title(r'{d[0]}$_\mathrm{{x}}${d[1]}$_\mathrm{{1-x}}$'.format(d=chempot_labels))
    elif isinstance(title, str) and title != '':
        ax.set_title(title)

    plt.locator_params(nbins=3)
    if hull._non_elemental:
        ax.set_xlabel(r'x in ({d[0]})$_\mathrm{{x}}$({d[1]})$_\mathrm{{1-x}}$'.format(d=chempot_labels))
    else:
        ax.set_xlabel(r'x in {d[0]}$_\mathrm{{x}}${d[1]}$_\mathrm{{1-x}}$'.format(d=chempot_labels))

    ax.grid(False)
    ax.set_xlim(-0.05, 1.05)
    ax.set_xticks([0, 0.25, 0.5, 0.75, 1])
    ax.set_xticklabels(ax.get_xticks())
    ax.set_ylabel('Formation energy (eV/atom)')

    if hull.savefig or any([kwargs.get(ext) for ext in SAVE_EXTS]):
        import os
        fname = plot_fname or ''.join(hull.species) + '_hull'
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                fname_tmp = fname
                ind = 0
                while os.path.isfile('{}.{}'.format(fname_tmp, ext)):
                    ind += 1
                    fname_tmp = fname + str(ind)

                fname = fname_tmp
                plt.savefig('{}.{}'.format(fname, ext),
                            bbox_inches='tight', transparent=True)
                print('Wrote {}.{}'.format(fname, ext))

    if show:
        plt.show()

    return ax


@plotting_function
def plot_ensemble_hull(hull, data_key,
                       ax=None,
                       formation_energy_key='formation_enthalpy_per_atom',
                       plot_points=False,
                       plot_hull_points=True,
                       alpha_scale=0.25,
                       plot_hulls=True,
                       voltages=False,
                       show=True,
                       plot_fname=None,
                       **kwargs):
    """ Plot and generate an ensemble of hulls. If axis not requested,
    a histogram of frequency of a particular concentration appearing on
    the convex hull is also generated as a second axis.

    Parameters:
        hull (QueryConvexHull): hull object created with a cursor that
            contains the data key for all entries in hull.cursor.
        data_key (str): the key under which ensemble data is stored.

    Keyword arguments:
        ax (matplotlib.axes.Axes): matplotlib axis object on which to plot.
        formation_energy_key (str): the key under which formation energies have been stored.
        alpha_scale (float): value by which to scale transparency of hulls.
        plot_points (bool): whether to plot the hull points for each hull in the ensemble.
        plot_hulls (bool): whether to plot the hull tie-lines for each hull in the ensemble.
        voltages (bool): compute average voltage and heatmaps.

    """
    import matplotlib.pyplot as plt

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    n_hulls = len(hull.phase_diagrams)
    plot_2d_hull(hull, ax=ax, plot_points=False, plot_hull_points=True, show=False, **kwargs)
    min_ef = 0
    colours_list = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])
    alpha = min([1, max([1/(alpha_scale*n_hulls), 0.01])])
    for ind, _ in enumerate(hull.phase_diagrams):
        hull_cursor = [doc for doc in hull.cursor if doc[data_key]['hull_distance'][ind] <= 0.0 + EPS]
        min_ef = np.min([doc[data_key][formation_energy_key][ind] for doc in hull_cursor] + [min_ef])
        if plot_hulls:
            ax.plot([doc['concentration'][0] for doc in hull_cursor],
                    [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                    alpha=alpha, c='k', lw=0.5, zorder=0)
        if plot_hull_points:
            ax.scatter([doc['concentration'][0] for doc in hull_cursor],
                       [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                       alpha=alpha, marker='o', c=colours_list[1], lw=0, zorder=0)
        if plot_points:
            ax.scatter([doc['concentration'][0] for doc in hull.cursor],
                       [doc[data_key][formation_energy_key][ind] for doc in hull.cursor],
                       alpha=alpha, marker='o', c='k', s=5, lw=0, zorder=0)

    ax.set_ylim(1.1*min_ef)

    if hull.savefig or any(kwargs.get(ext) for ext in SAVE_EXTS):
        if plot_fname is not None:
            fname = plot_fname
        else:
            fname = ''.join(hull.species) + data_key + '_hull'
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                plt.savefig('{}.{}'.format(fname, ext),
                            bbox_inches='tight', transparent=True)
                print('Wrote {}.{}'.format(fname, ext))


@plotting_function
def plot_temperature_hull(
    hull,
    cmap='plasma',
    cmap_limits=(0.2, 0.8),
    ax=None,
    formation_energy_key='formation_free_energy_per_atom',
    plot_points=False,
    show_cbar=True,
    plot_hull_points=True,
    alpha_scale=1,
    lw_scale=1,
    plot_hulls=True,
    voltages=False,
    show=True,
    plot_fname=None,
    **kwargs
):
    """ Plot and generate an ensemble of hulls. If axis not requested,
    a histogram of frequency of a particular concentration appearing on
    the convex hull is also generated as a second axis.

    Parameters:
        hull (QueryConvexHull): hull object created with a cursor that
            contains the data key for all entries in hull.cursor.

    Keyword arguments:
        ax (matplotlib.axes.Axes): matplotlib axis object on which to plot.
        formation_energy_key (str): the key under which formation energies have been stored.
        alpha_scale (float): value by which to scale transparency of hulls.
        plot_points (bool): whether to plot the hull points for each hull in the ensemble.
        plot_hulls (bool): whether to plot the hull tie-lines for each hull in the ensemble.
        voltages (bool): compute average voltage and heatmaps.

    """
    import matplotlib.pyplot as plt
    from matplotlib.colors import rgb2hex
    data_key = hull.data_key

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    n_hulls = len(hull.phase_diagrams)
    colours = [
        rgb2hex(col) for col in
        plt.cm.get_cmap(cmap)(np.linspace(*cmap_limits, n_hulls)).tolist()
    ]

    min_ef = 0
    alpha = alpha_scale

    # hack the energy key so that labels work
    _cached_key = hull.energy_key
    hull.energy_key = hull.chempot_energy_key
    max_temperature = max(hull.temperatures)
    # set up initial plot without plotting tie line
    ax = plot_2d_hull(
        hull, ax=ax,
        plot_tie_line=False,
        plot_points=False,
        plot_hull_points=False,
        show=False,
        **kwargs
    )
    hull.energy_key = _cached_key

    ax.plot([doc['concentration'][0] for doc in hull.hull_cursor],
            [doc['formation_' + hull.chempot_energy_key] for doc in hull.hull_cursor],
            marker='o',
            alpha=1, c='k', lw=2*lw_scale, ls='--', label='Static', zorder=1e5)

    ind = 0
    hull_cursor = [doc for doc in hull.cursor if doc[data_key]['hull_distance'][ind] <= 0.0 + EPS]
    ax.plot([doc['concentration'][0] for doc in hull_cursor],
            [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
            marker='o', c=colours[ind], lw=2*lw_scale,
            markeredgewidth=1.5, markeredgecolor='k',
            ls='--', zorder=1e5, label='Static + ZPE')

    # plot remaining temperatures
    for ind, _ in enumerate(hull.temperatures[1:]):
        hull_cursor = [doc for doc in hull.cursor if doc[data_key]['hull_distance'][ind] <= 0.0 + EPS]
        min_ef = np.min([doc[data_key][formation_energy_key][ind] for doc in hull_cursor] + [min_ef])
        colour_ind = int(n_hulls * hull.temperatures[ind] / max_temperature)
        if plot_hulls:
            ax.plot([doc['concentration'][0] for doc in hull_cursor],
                    [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                    alpha=alpha, c=colours[colour_ind], lw=1*lw_scale, zorder=0)
        if plot_hull_points:
            ax.scatter([doc['concentration'][0] for doc in hull_cursor],
                       [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                       alpha=1, marker='o', c=colours[colour_ind], lw=0.5,
                       edgecolor='k', label="Structures on hull" if ind == n_hulls - 2 else None,
                       zorder=1e4)
        if plot_points:
            ax.scatter([doc['concentration'][0] for doc in hull.cursor],
                       [doc[data_key][formation_energy_key][ind] for doc in hull.cursor],
                       label="Structures above hull" if ind == n_hulls - 2 else None,
                       alpha=1, marker='o', edgecolor='w', c=colours[colour_ind], lw=0.5, zorder=1e-3)

    ax.set_ylim(1.1*min_ef)

    if show_cbar:
        import matplotlib.colors
        mappable = plt.cm.ScalarMappable(
            cmap=matplotlib.colors.LinearSegmentedColormap.from_list('cut', colours),
            norm=plt.Normalize(vmin=0, vmax=np.max(hull.temperatures))
        )
        mappable._A = hull.temperatures
        cbar = plt.colorbar(mappable, alpha=alpha)
        cbar.ax.tick_params(length=0)
        cbar.ax.yaxis.set_ticks_position('right')
        cbar.ax.set_frame_on(False)
        cbar.outline.set_visible(False)
        cbar.set_label('Temperature (K)')

    if hull.savefig or any(kwargs.get(ext) for ext in SAVE_EXTS):
        if plot_fname is not None:
            fname = plot_fname
        else:
            fname = ''.join(hull.species) + data_key + '_hull'
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                plt.savefig('{}.{}'.format(fname, ext),
                            bbox_inches='tight', transparent=True)
                print('Wrote {}.{}'.format(fname, ext))

    ax.legend()

    return ax


@plotting_function
def plot_ternary_hull(hull, axis=None, show=True, plot_points=True, hull_cutoff=None, fig_height=None,
                      label_cutoff=None, label_corners=True, expecting_cbar=True, labels=None, plot_fname=None,
                      efmap=None, sampmap=None, capmap=None, pathways=False, **kwargs):
    """ Plot calculated ternary hull as a 2D projection.

    Parameters:
        hull (matador.hull.QueryConvexHull): matador hull object.

    Keyword arguments:
        axis (matplotlib.axes.Axes): matplotlib axis object on which to plot.
        show (bool): whether or not to show plot in X window.
        plot_points (bool): whether or not to plot each structure as a point.
        label_cutoff (float/:obj:`tuple` of :obj:`float`): draw labels less than or
            between these distances form the hull, also read from hull.args.
        expecting_cbar (bool): whether or not to space out the plot to preserve
            aspect ratio if a colourbar is present.
        labels (bool): whether or not to label on-hull structures
        label_corners (bool): whether or not to put axis labels on corners or edges.
        png/pdf/svg (bool): whether or not to write the plot to a file.
        plot_fname (str): filename to write plot to.
        efmap (bool): plot heatmap of formation energy,
        sampmap (bool): plot heatmap showing sampling density,
        capmap (bool): plot heatmap showing gravimetric capacity.
        pathways (bool): plot the pathway from the starting electrode to active ion.

    Returns:
        matplotlib.axes.Axes: matplotlib axis with plot.

    """
    import ternary
    import matplotlib.pyplot as plt
    import matplotlib.colors as colours
    from matador.utils.chem_utils import get_generic_grav_capacity

    plt.rcParams['axes.linewidth'] = 0
    plt.rcParams['xtick.major.size'] = 0
    plt.rcParams['ytick.major.size'] = 0
    plt.rcParams['xtick.minor.size'] = 0
    plt.rcParams['ytick.minor.size'] = 0

    if efmap is None:
        efmap = hull.args.get('efmap')
    if sampmap is None:
        sampmap = hull.args.get('sampmap')
    if capmap is None:
        capmap = hull.args.get('capmap')
    if pathways is None:
        pathways = hull.args.get('pathways')

    if labels is None:
        labels = hull.args.get('labels')
    if label_cutoff is None:
        label_cutoff = hull.args.get('label_cutoff')
        if label_cutoff is None:
            label_cutoff = 0
    else:
        labels = True

    if hull_cutoff is None and hull.hull_cutoff is None:
        hull_cutoff = 0
    else:
        hull_cutoff = hull.hull_cutoff

    print('Plotting ternary hull...')
    if capmap or efmap:
        scale = 100
    elif sampmap:
        scale = 20
    else:
        scale = 1

    if axis is not None:
        fig, ax = ternary.figure(scale=scale, ax=axis)
    else:
        fig, ax = ternary.figure(scale=scale)

    # maintain aspect ratio of triangle
    if fig_height is None:
        _user_height = plt.rcParams.get("figure.figsize", (8, 6))[0]
    else:
        _user_height = fig_height
    if capmap or efmap or sampmap:
        fig.set_size_inches(_user_height, 5/8 * _user_height)
    elif not expecting_cbar:
        fig.set_size_inches(_user_height, _user_height)
    else:
        fig.set_size_inches(_user_height, 5/6.67 * _user_height)

    ax.boundary(linewidth=2.0, zorder=99)
    ax.clear_matplotlib_ticks()

    chempot_labels = [get_formula_from_stoich(
        get_stoich_from_formula(species, sort=False), sort=False, tex=True) for species in hull.species
    ]

    ax.gridlines(color='black', multiple=scale * 0.1, linewidth=0.5)
    ticks = [float(val) for val in np.linspace(0, 1, 6)]
    if label_corners:
        # remove 0 and 1 ticks when labelling corners
        ticks = ticks[1:-1]
        ax.left_corner_label(chempot_labels[2], fontsize='large')
        ax.right_corner_label(chempot_labels[0], fontsize='large')
        ax.top_corner_label(chempot_labels[1], fontsize='large', offset=0.16)
    else:
        ax.left_axis_label(chempot_labels[2], fontsize='large', offset=0.12)
        ax.right_axis_label(chempot_labels[1], fontsize='large', offset=0.12)
        ax.bottom_axis_label(chempot_labels[0], fontsize='large', offset=0.08)
        ax.set_title('-'.join(['{}'.format(label) for label in chempot_labels]), fontsize='large', y=1.02)

    ax.ticks(axis='lbr', linewidth=1, offset=0.025, fontsize='small',
             locations=(scale * np.asarray(ticks)).tolist(),
             ticks=ticks, tick_formats='%.1f')

    concs = np.zeros((len(hull.structures), 3))
    concs[:, :-1] = hull.structures[:, :-1]
    for i in range(len(concs)):
        # set third triangular coordinate
        concs[i, -1] = 1 - concs[i, 0] - concs[i, 1]

    stable = concs[np.where(hull.hull_dist <= 0 + EPS)]

    # sort by hull distances so things are plotting the right order
    concs = concs[np.argsort(hull.hull_dist)].tolist()
    hull_dist = np.sort(hull.hull_dist)

    filtered_concs = []
    filtered_hull_dists = []
    for ind, conc in enumerate(concs):
        if conc not in filtered_concs:
            if hull_dist[ind] <= hull.hull_cutoff or (hull.hull_cutoff == 0 and hull_dist[ind] < 0.1):
                filtered_concs.append(conc)
                filtered_hull_dists.append(hull_dist[ind])
    if hull.args.get('debug'):
        print('Trying to plot {} points...'.format(len(filtered_concs)))

    concs = np.asarray(filtered_concs)
    hull_dist = np.asarray(filtered_hull_dists)

    min_cut = 0.0
    max_cut = 0.2

    hull.colours = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])
    hull.default_cmap_list = get_linear_cmap(hull.colours[1:4], list_only=True)
    hull.default_cmap = get_linear_cmap(hull.colours[1:4], list_only=False)
    n_colours = len(hull.default_cmap_list)
    colours_hull = hull.default_cmap_list

    cmap = hull.default_cmap
    cmap_full = plt.cm.get_cmap('Pastel2')
    pastel_cmap = colours.LinearSegmentedColormap.from_list('Pastel2', cmap_full.colors)

    for plane in hull.convex_hull.planes:
        plane.append(plane[0])
        plane = np.asarray(plane)
        ax.plot(scale * plane, c=hull.colours[0], lw=1.5, alpha=1, zorder=98)

    if pathways:
        for phase in stable:
            if phase[0] == 0 and phase[1] != 0 and phase[2] != 0:
                ax.plot([scale * phase, [scale, 0, 0]], c='r', alpha=0.2, lw=6, zorder=99)

    # add points
    if plot_points:
        colours_list = []
        colour_metric = hull_dist
        for i, _ in enumerate(colour_metric):
            if colour_metric[i] >= max_cut:
                colours_list.append(n_colours - 1)
            elif colour_metric[i] <= min_cut:
                colours_list.append(0)
            else:
                colours_list.append(int((n_colours - 1) * (colour_metric[i] / max_cut)))
        colours_list = np.asarray(colours_list)
        ax.scatter(scale*stable, marker='o', color=hull.colours[1], edgecolors='black', zorder=9999999,
                   s=150, lw=1.5)
        ax.scatter(scale*concs, colormap=cmap, colorbar=True, cbarlabel='Distance from hull (eV/atom)',
                   c=colour_metric, vmax=max_cut, vmin=min_cut, zorder=1000, s=40, alpha=0)
        for i, _ in enumerate(concs):
            ax.scatter(
                scale * concs[i].reshape(1, 3),
                color=colours_hull[colours_list[i]],
                marker='o',
                zorder=10000 - colours_list[i],
                s=70 * (1 - float(colours_list[i]) / n_colours) + 15,
                lw=1,
                edgecolors='black'
            )

    # add colourmaps
    if capmap:
        capacities = dict()
        from ternary.helpers import simplex_iterator
        for (i, j, k) in simplex_iterator(scale):
            capacities[(i, j, k)] = get_generic_grav_capacity([
                float(i) / scale, float(j) / scale, float(scale - i - j) / scale
            ], hull.species)
        ax.heatmap(capacities, style="hexagonal", cbarlabel='Gravimetric capacity (mAh/g)',
                   vmin=0, vmax=3000, cmap=pastel_cmap)
    elif efmap:
        energies = dict()
        fake_structures = []
        from ternary.helpers import simplex_iterator
        for (i, j, k) in simplex_iterator(scale):
            fake_structures.append([float(i) / scale, float(j) / scale, 0.0])
        fake_structures = np.asarray(fake_structures)
        plane_energies = hull.get_hull_distances(fake_structures, precompute=False)
        ind = 0
        for (i, j, k) in simplex_iterator(scale):
            energies[(i, j, k)] = -1 * plane_energies[ind]
            ind += 1
        if isinstance(efmap, str):
            efmap = efmap
        else:
            efmap = 'BuPu_r'
        ax.heatmap(energies, style="hexagonal", cbarlabel='Formation energy (eV/atom)', vmax=0, cmap=efmap)
    elif sampmap:
        sampling = dict()
        from ternary.helpers import simplex_iterator
        eps = 1.0 / float(scale)
        for (i, j, k) in simplex_iterator(scale):
            sampling[(i, j, k)] = np.size(np.where((concs[:, 0] <= float(i)/scale + eps) *
                                                   (concs[:, 0] >= float(i)/scale - eps) *
                                                   (concs[:, 1] <= float(j)/scale + eps) *
                                                   (concs[:, 1] >= float(j)/scale - eps) *
                                                   (concs[:, 2] <= float(k)/scale + eps) *
                                                   (concs[:, 2] >= float(k)/scale - eps)))
        ax.heatmap(sampling, style="hexagonal", cbarlabel='Number of structures', cmap='afmhot')

    # add labels
    if labels:
        label_cursor = _get_hull_labels(hull, label_cutoff=label_cutoff)
        if len(label_cursor) == 1:
            label_coords = [[0.25, 0.5]]
        else:
            label_coords = [[0.1+(val-0.5)*0.3, val] for val in np.linspace(0.5, 0.8, int(round(len(label_cursor)/2.)+1))]
            label_coords += [[0.9-(val-0.5)*0.3, val+0.2] for val in np.linspace(0.5, 0.8, int(round(len(label_cursor)/2.)))]
        from matador.utils.hull_utils import barycentric2cart
        for ind, doc in enumerate(label_cursor):
            conc = np.asarray(doc['concentration'] + [1 - sum(doc['concentration'])])
            formula = get_formula_from_stoich(
                doc['stoichiometry'], sort=False, tex=True, latex_sub_style=r'\mathregular', elements=hull.species
            )
            arrowprops = dict(arrowstyle="-|>", color='k', lw=2, alpha=0.5, zorder=1, shrinkA=2, shrinkB=4)
            cart = barycentric2cart([doc['concentration'] + [0]])[0][:2]
            min_dist = 1e20
            closest_label = 0
            for coord_ind, coord in enumerate(label_coords):
                dist = np.sqrt((cart[0] - coord[0])**2 + (cart[1] - coord[1])**2)
                if dist < min_dist:
                    min_dist = dist
                    closest_label = coord_ind
            ax.annotate(formula, scale*conc,
                        textcoords='data',
                        xytext=[scale*val for val in label_coords[closest_label]],
                        ha='right',
                        va='bottom',
                        arrowprops=arrowprops)
            del label_coords[closest_label]

    plt.tight_layout(w_pad=0.2)
    # important for retaining labels if exporting to PDF
    # see https://github.com/marcharper/python-ternary/issues/36
    ax._redraw_labels() # noqa

    if hull.savefig:
        fname = plot_fname or ''.join(hull.species) + '_hull'
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                plt.savefig('{}.{}'.format(fname, ext),
                            bbox_inches='tight', transparent=True)
                print('Wrote {}.{}'.format(fname, ext))
    elif show:
        print('Showing plot...')
        plt.show()

    return ax


def _scatter_plot_by_source(hull, ax, scale, kwargs,
                            sources=None, source_labels=None, plot_hull_points=True, legend_kwargs=None):
    """ Add scatter points to the hull depending on the guessed
    provenance of a structure.

    """
    from matador.utils.cursor_utils import get_guess_doc_provenance
    if sources is None:
        sources = ['AIRSS', 'GA', 'OQMD', 'SWAPS', 'ICSD', 'DOI', 'SM', 'MP', 'PF', 'Other']
    if source_labels is None:
        source_labels = sources
    else:
        assert len(source_labels) == len(sources)

    # hack: double length of hull colours
    hull.colours.extend(hull.colours)

    colour_choices = {source: hull.colours[ind + 1] for ind, source in enumerate(sources)}
    colours = []
    concs = []
    energies = []
    zorders = []
    for doc in hull.cursor:
        source = get_guess_doc_provenance(doc['source'])
        if source not in sources:
            # use grey for undesired sources
            colours.append(hull.colours[-2])
            source = 'Other'
            if 'Other' not in sources:
                sources.append('Other')
                source_labels.append('Other')
                colour_choices['Other'] = hull.colours[-2]
        else:
            colours.append(source)
        zorders.append(sources.index(source))
        concs.append(doc['concentration'])
        energies.append(doc['formation_{}'.format(hull.energy_key)])

    sources_present = set(colours)
    sources_present = [source for source in sources if source in sources_present]
    colour_choices = {source: hull.colours[ind + 1] for ind, source in enumerate(sources_present)}
    colours = [colour_choices[src] for src in colours]
    alpha = kwargs.get('alpha')
    if alpha is None:
        alpha = 0.2

    for ind, conc in enumerate(concs):
        if hull.cursor[ind]['hull_distance'] <= 0 + 1e-9 and not plot_hull_points:
            ax.scatter(conc, energies[ind],
                       c=colours[ind], alpha=1, s=scale*40, edgecolor='k',
                       zorder=zorders[ind]+1e5, lw=1.5)
        else:
            ax.scatter(conc, energies[ind],
                       c=colours[ind], alpha=alpha, s=scale*20,
                       zorder=zorders[ind]+100)

    for ind, source in enumerate(sources_present):
        ax.scatter(1e10, 1e10, c=colour_choices[source], label=source_labels[ind], alpha=alpha, lw=1)

    if legend_kwargs is not None:
        legend = ax.legend(**legend_kwargs)
    else:
        legend = ax.legend(ncol=2)
    legend.set_zorder(1e20)

    return ax
